"""Exception classes for architect_ai package."""

from .tool_exceptions import (
    ToolError,
    PrecallableToolNotFoundError,
    PrecallableToolExecutionError,
    PrecallableToolRuntimeError,
)

from .architect_exceptions import (
    ArchitectError,
    PrecallableToolError,
    BuildPlanParsingError,
    BlueprintNotFoundError,
    ToolExecutionModeError,
    ToolExecutionError,
    ReferenceResolutionError,
    InvalidReferencePathError,
    StageValidationError,
    BuildPlanGenerationError,
    ToolPicklingError,
    ToolNotFoundError,
    ExecutableFailedToStartError,
    BuildGenerationError,
)

from .toolbox_exceptions import (
    ToolBoxError,
    ToolValidationError,
    DuplicateToolError,
)

__all__ = [
    # Tool exceptions
    "ToolError",
    "PrecallableToolNotFoundError",
    "PrecallableToolExecutionError",
    "PrecallableToolRuntimeError",
    
    # Architect exceptions
    "ArchitectError",
    "PrecallableToolError",
    "BuildPlanParsingError",
    "BlueprintNotFoundError",
    "ToolExecutionModeError",
    "ToolExecutionError",
    "ReferenceResolutionError",
    "InvalidReferencePathError",
    "StageValidationError",
    "BuildPlanGenerationError",
    "ToolPicklingError",
    "ToolNotFoundError",
    "ExecutableFailedToStartError",
    "BuildGenerationError",
    
    # Toolbox exceptions
    "ToolBoxError",
    "ToolValidationError",
    "DuplicateToolError",
]



from typing import Dict, List, Optional

from .tool import Tool


class ToolBox:
    def __init__(self, tools: Optional[List[Tool]] = None):
        """
        Initialize the ToolBox as a container for tools available to the Architect.

        Args:
            tools (Optional[List[Tool]]): A list of Tool objects to add to the ToolBox
            
        Returns:
            None: Initializes the toolbox instance
        """
        self.tools: Dict[str, Tool] = {}
        if tools:
            self._validate_and_add_tools(tools)

    def add_tool(self, tool: Tool) -> None:
        """
        Adds a Tool object to the ToolBox.

        Args:
            tool (Tool): The Tool object to add to the ToolBox
            
        Returns:
            None: Adds the tool to the toolbox
            
        Raises:
            ValueError: If tool is missing name attribute or name already exists
        """
        if not hasattr(tool, "name"):
            raise ValueError("Tool must have a name attribute")
        if tool.name in self.tools:
            raise ValueError(f"Tool with name '{tool.name}' already exists")
        self.tools[tool.name] = tool

    def set_tool_list(self, tools: List[Tool]) -> None:
        """
        Sets the list of Tool objects to the ToolBox, removing any existing tools.

        Args:
            tools (List[Tool]): A list of Tool objects to set in the ToolBox
            
        Returns:
            None: Replaces all existing tools with the provided list
        """
        self.tools = {}
        self._validate_and_add_tools(tools)

    def find_by_name(self, name: str) -> Optional[Tool]:
        """
        Finds a Tool object by its name.

        Args:
            name (str): The name of the Tool object to find

        Returns:
            Optional[Tool]: The Tool object if found, otherwise None
        """
        return self.tools.get(name)

    def remove_tool(self, name: str) -> None:
        """
        Removes a Tool object from the ToolBox.

        Args:
            name (str): The name of the Tool object to remove
            
        Returns:
            None: Removes the tool from the toolbox if it exists
        """
        if name in self.tools:
            del self.tools[name]

    def open_tool(self, name: str) -> Optional[str]:
        """
        Creates a string representation of a Tool object formatted for consumption by a Large Language Model.

        Args:
            name (str): The name of the Tool object to open

        Returns:
            Optional[str]: A string representation of the Tool object, or None if not found
        """
        tool: Optional[Tool] = self.tools.get(name)
        if not tool:
            return None
        tool_details: str = f"TOOL NAME: {tool.name}\n"
        tool_details += f"Purpose: {tool.purpose}\n"
        tool_details += "Required Parameters:\n"
        for param_name, (data_type, description) in tool.required_parameters.items():
            tool_details += f"  {param_name} ({data_type}): {description}\n"
        tool_details += "Optional Parameters:\n"
        for param_name, (data_type, description) in tool.optional_parameters.items():
            tool_details += f"  {param_name} ({data_type}): {description}\n"
        tool_details += "\nOutput Descriptions:\n"
        for output_name, (data_type, description) in tool.output_descriptions.items():
            tool_details += f"  {output_name} ({data_type}): {description}\n"
        return tool_details

    def open_all_tools(self) -> str:
        """
        Opens each tool and appends the details to a single string.

        Args:
            None
            
        Returns:
            str: A string representation of all the Tool objects in the ToolBox
        """
        all_tool_details: str = "TOOLS:\n--------------------------------\n"
        for name in self.tools.keys():
            tool_details: Optional[str] = self.open_tool(name)
            if tool_details:
                all_tool_details += tool_details + "\n"
        return all_tool_details

    def _validate_and_add_tools(self, tools: List[Tool]) -> None:
        """
        Validates the tools and adds them to the ToolBox.

        Args:
            tools (List[Tool]): A list of Tool objects to add to the ToolBox
            
        Returns:
            None: Validates and adds tools to the toolbox
            
        Raises:
            ValueError: If tool validation fails or duplicate names exist
        """
        for tool in tools:
            if not hasattr(tool, "name"):
                raise ValueError("Tool must have a name attribute")
            if tool.name in self.tools:
                raise ValueError(
                    f"Tool with name '{tool.name}' already exists"
                )
            self.tools[tool.name] = tool



"""Tool interfaces for architect_ai package."""

from abc import ABC, abstractmethod
from typing import Dict, Tuple, Any, Optional, List, Callable, TYPE_CHECKING
from .execution_mode import ExecutionMode

if TYPE_CHECKING:
    from .concurrent_executable import ConcurrentExecutable

def wrap_tool_func(func: Callable, parameters: Dict[str, Any], concurrent_executables: Optional[List["ConcurrentExecutable"]] = None) -> Callable:
    """
    Wrap a tool function with predefined parameters and concurrent executables for deferred execution.
    
    Args:
        func (Callable): The tool function to wrap
        parameters (Dict[str, Any]): Parameters to pass to the function
        concurrent_executables (Optional[List[ConcurrentExecutable]]): Available concurrent executables
        
    Returns:
        Callable: Wrapped function that can be called without parameters
    """
    def wrapped_func():
        return func(parameters, concurrent_executables)
    return wrapped_func

class Tool(ABC):
    """Tool that runs during build plan execution with parameters and precallable results."""

    @abstractmethod
    def use(self, parameters: Dict[str, Any], concurrent_executables: Optional[List["ConcurrentExecutable"]] = None) -> Any:
        """
        Execute the tool with parameters and concurrent executables.

        Args:
            parameters (Dict[str, Any]): Tool parameters from the build plan
            concurrent_executables (Optional[List[ConcurrentExecutable]]): Available concurrent executables
        
        Returns:
            Any: The result of the tool execution
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def execution_mode(self) -> ExecutionMode:
        """
        Defines how this tool should be executed.
        
        Args:
            None
            
        Returns:
            ExecutionMode: Execution mode for this tool (ASYNCIO for I/O-bound, THREAD for CPU-bound, PROCESS for heavy CPU-bound)
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def name(self) -> str:
        """
        Unique identifier for this tool.
        
        Args:
            None
            
        Returns:
            str: The unique name of this tool
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def purpose(self) -> str:
        """
        Description of what this tool accomplishes.
        
        Args:
            None
            
        Returns:
            str: Purpose description of what this tool does
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def required_parameters(self) -> Dict[str, Tuple[str, str]]:
        """
        Map of required parameter names to data type and description tuples.
        
        Args:
            None
            
        Returns:
            Dict[str, Tuple[str, str]]: Map of required parameter names to (DataType, description) tuples
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def optional_parameters(self) -> Dict[str, Tuple[str, str]]:
        """
        Map of optional parameter names to data type and description tuples.
        
        Args:
            None
            
        Returns:
            Dict[str, Tuple[str, str]]: Map of optional parameter names to (DataType, description) tuples
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def output_descriptions(self) -> Dict[str, Tuple[str, str]]:
        """
        Map of output names to data type and description tuples.
        
        Args:
            None
            
        Returns:
            Dict[str, Tuple[str, str]]: Map of output names to (DataType, description) tuples
        """
        raise NotImplementedError




"""Toolbox-related exception classes for architect_ai package."""

from typing import Optional


class ToolBoxError(Exception):
    """Base exception for ToolBox-related errors."""
    
    def __init__(self, message: str, **context):
        """
        Initialize the toolbox error with message and context.
        
        Args:
            message (str): Error message
            **context: Additional context information
            
        Returns:
            None: Initializes the toolbox error instance
        """
        super().__init__(message)
        self.context = context


class ToolValidationError(ToolBoxError):
    """Raised when tool validation fails."""
    
    def __init__(self, issue: str, tool_type: Optional[str] = None, tool_name: Optional[str] = None, **context):
        """
        Initialize the tool validation error.
        
        Args:
            issue (str): Description of the validation issue
            tool_type (Optional[str]): Type of tool that failed validation
            tool_name (Optional[str]): Name of tool that failed validation
            **context: Additional context information
            
        Returns:
            None: Initializes the tool validation error instance
        """
        super().__init__(
            f"Tool validation failed: {issue}",
            issue=issue,
            tool_type=tool_type,
            tool_name=tool_name,
            **context
        )


class DuplicateToolError(ToolBoxError):
    """Raised when attempting to add a tool with a name that already exists."""
    
    def __init__(self, tool_name: str, **context):
        """
        Initialize the duplicate tool error.
        
        Args:
            tool_name (str): Name of the duplicate tool
            **context: Additional context information
            
        Returns:
            None: Initializes the duplicate tool error instance
        """
        super().__init__(
            f"Tool with name '{tool_name}' already exists",
            tool_name=tool_name,
            **context
        )



import asyncio
from typing import Any, Callable, Optional, Dict, List
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
import logging
from .tool import Tool
from .execution_mode import ExecutionMode

logger = logging.getLogger(__name__)


class ConcurrentExecutable:
    """
    Unified wrapper for handling asyncio tasks, threading operations, and multiprocessing operations.

    Provides a consistent API regardless of execution type.
    """

    def __init__(
        self,
        name: str,
        execution_mode: ExecutionMode,
        func: Callable,
        func_args: Optional[Dict[str, Any]] = None,
        func_initial_concurrent_executables: Optional[
            List["ConcurrentExecutable"]
        ] = None,
        asyncio_event_loop: Optional[asyncio.AbstractEventLoop] = None,
        thread_pool_executor: Optional[ThreadPoolExecutor] = None,
        process_pool_executor: Optional[ProcessPoolExecutor] = None,
    ):
        """
        Initialize a concurrent executable wrapper for different execution modes.
        
        Args:
            name (str): Name identifier for this executable
            execution_mode (ExecutionMode): Mode of execution (ASYNCIO, THREAD, PROCESS, IMMEDIATE)
            func (Callable): Function to execute
            func_args (Optional[Dict[str, Any]]): Arguments to pass to the function
            func_initial_concurrent_executables (Optional[List[ConcurrentExecutable]]): Initial concurrent executables
            asyncio_event_loop (Optional[asyncio.AbstractEventLoop]): Event loop for asyncio execution
            thread_pool_executor (Optional[ThreadPoolExecutor]): Thread pool for thread execution
            process_pool_executor (Optional[ProcessPoolExecutor]): Process pool for process execution
            
        Returns:
            None: Initializes the concurrent executable instance
        """
        self.name = name
        self.execution_mode = execution_mode
        self.func = func
        self.func_args = func_args
        self.func_initial_concurrent_executables = func_initial_concurrent_executables
        self.asyncio_event_loop = asyncio_event_loop
        self.thread_pool_executor = thread_pool_executor
        self.process_pool_executor = process_pool_executor
        self.running_future: Any = None
        self.validate_initialization()
        self.is_current_running = False

    @classmethod
    def from_tool(
        cls,
        tool: Tool,
        parameters: Dict[str, Any],
        initial_concurrent_executables: Optional[List["ConcurrentExecutable"]] = None,
        asyncio_event_loop: Optional[asyncio.AbstractEventLoop] = None,
        thread_pool_executor: Optional[ThreadPoolExecutor] = None,
        process_pool_executor: Optional[ProcessPoolExecutor] = None,
    ) -> "ConcurrentExecutable":
        """
        Create a ConcurrentExecutable from a Tool instance.
        
        Args:
            tool (Tool): Tool instance to wrap
            parameters (Dict[str, Any]): Parameters to pass to the tool
            initial_concurrent_executables (Optional[List[ConcurrentExecutable]]): Initial concurrent executables
            asyncio_event_loop (Optional[asyncio.AbstractEventLoop]): Event loop for asyncio execution
            thread_pool_executor (Optional[ThreadPoolExecutor]): Thread pool for thread execution
            process_pool_executor (Optional[ProcessPoolExecutor]): Process pool for process execution
            
        Returns:
            ConcurrentExecutable: New instance wrapping the tool
        """
        return cls(
            name=tool.name,
            execution_mode=tool.execution_mode,
            func=tool.use,
            func_args=parameters,
            func_initial_concurrent_executables=initial_concurrent_executables,
            asyncio_event_loop=asyncio_event_loop,
            thread_pool_executor=thread_pool_executor,
            process_pool_executor=process_pool_executor,
        )

    def validate_initialization(self) -> None:
        """
        Validate that the required executors are provided for the execution mode.
        
        Args:
            None
            
        Returns:
            None: Validates initialization parameters
            
        Raises:
            ValueError: If required executor is missing for the execution mode
        """
        if self.execution_mode == ExecutionMode.ASYNCIO:
            if self.asyncio_event_loop is None:
                raise ValueError("Asyncio event loop is required for asyncio execution")
        elif self.execution_mode == ExecutionMode.THREAD:
            if self.thread_pool_executor is None:
                raise ValueError(
                    "Thread pool executor is required for thread execution"
                )
        elif self.execution_mode == ExecutionMode.PROCESS:
            if self.process_pool_executor is None:
                raise ValueError(
                    "Process pool executor is required for process execution"
                )

    def start(self) -> None:
        """
        Start the concurrent executable based on its execution mode.
        
        Args:
            None
            
        Returns:
            None: Starts the executable and sets running state
            
        Raises:
            ValueError: If already running or invalid execution mode or missing executor
        """
        if self.is_current_running:
            raise ValueError("Concurrent executable is already running")
        try:
            logger.debug(f"Starting ConcurrentExecutable '{self.name}' with mode {self.execution_mode}")
            if self.execution_mode == ExecutionMode.ASYNCIO:
                if self.asyncio_event_loop is None:
                    raise ValueError("Asyncio event loop is required for asyncio execution")
                if self.func_args is None:
                    self.running_future = asyncio.run_coroutine_threadsafe(
                        self.func(), self.asyncio_event_loop
                    )
                else:
                    self.running_future = asyncio.run_coroutine_threadsafe(
                        self.func(self.func_args, self.func_initial_concurrent_executables),
                        self.asyncio_event_loop,
                    )
            elif self.execution_mode == ExecutionMode.THREAD:
                if self.thread_pool_executor is None:
                    raise ValueError(
                        "Thread pool executor is required for thread execution"
                    )
                if self.func_args is None:
                    self.running_future = self.thread_pool_executor.submit(self.func)
                else:
                    self.running_future = self.thread_pool_executor.submit(
                        self.func, self.func_args, self.func_initial_concurrent_executables
                    )
            elif self.execution_mode == ExecutionMode.PROCESS:
                if self.process_pool_executor is None:
                    raise ValueError(
                        "Process pool executor is required for process execution"
                    )
                if self.func_args is None:
                    self.running_future = self.process_pool_executor.submit(self.func)
                else:
                    self.running_future = self.process_pool_executor.submit(
                        self.func, self.func_args
                    )
            elif self.execution_mode == ExecutionMode.IMMEDIATE:
                if self.func_args is None:
                    self.running_future = self.func()
                else:
                    self.running_future = self.func(self.func_args, self.func_initial_concurrent_executables)
            else:
                raise ValueError(f"Invalid execution mode: {self.execution_mode}")
            self.is_current_running = True
            logger.debug(f"ConcurrentExecutable '{self.name}' started successfully")
        except Exception as start_error:
            logger.warning(f"Failed to start ConcurrentExecutable '{self.name}': {type(start_error).__name__}: {start_error}")
            logger.warning(f"Full start failure traceback for '{self.name}':", exc_info=True)
            self.is_current_running = False
            raise

    def is_ready(self) -> bool:
        """
        Check if the concurrent executable is ready with computation complete.
        
        Args:
            None
            
        Returns:
            bool: True if computation is complete and results are ready, False otherwise
        """
        if not self.is_current_running:
            return False
        if self.execution_mode == ExecutionMode.IMMEDIATE:
            return True
        elif self.execution_mode in [ExecutionMode.THREAD, ExecutionMode.PROCESS, ExecutionMode.ASYNCIO]:
            if self.running_future is not None:
                return self.running_future.done()
            return False
        else:
            return False

    def fetch_results(self) -> Any:
        """
        Fetch the results from the completed concurrent executable.
        
        Args:
            None
            
        Returns:
            Any: Results from the completed execution
            
        Raises:
            ValueError: If executable is not running or execution fails
        """
        if not self.is_current_running:
            raise ValueError("Concurrent executable is not running")
        try:
            logger.debug(
                f"Fetching results for '{self.name}' in mode {self.execution_mode}; has_future={self.running_future is not None}"
            )
            if (
                self.execution_mode
                in [ExecutionMode.THREAD, ExecutionMode.PROCESS, ExecutionMode.ASYNCIO]
                and self.running_future is not None
            ):
                logger.debug(f"Waiting on future.result() for '{self.name}'")
                return self.running_future.result(timeout=60)
            elif self.execution_mode == ExecutionMode.IMMEDIATE:
                return self.running_future
            else:
                raise ValueError(
                    f"Invalid execution mode: {self.execution_mode} or no available task/future"
                )
        except Exception as fetch_error:
            logger.error(
                f"ConcurrentExecutable '{self.name}' failed with exception: {type(fetch_error).__name__}: {fetch_error}"
            )
            logger.error("Full traceback:", exc_info=True)
            raise ValueError(f"Failed to fetch results: {fetch_error}") from fetch_error

    async def fetch_results_async(self) -> Any:
        """
        Asynchronously fetch results without blocking the event loop.
        
        Args:
            None
            
        Returns:
            Any: Results from the completed execution
        """
        if not self.is_current_running:
            raise ValueError("Concurrent executable is not running")
        if self.execution_mode in [ExecutionMode.THREAD, ExecutionMode.PROCESS, ExecutionMode.ASYNCIO] and self.running_future is not None:
            wrapped = asyncio.wrap_future(self.running_future)
            return await asyncio.wait_for(wrapped, timeout=60)
        if self.execution_mode == ExecutionMode.IMMEDIATE:
            return self.running_future
        raise ValueError(f"Invalid execution mode: {self.execution_mode} or no available task/future")

    def stop(self) -> None:
        """
        Stop the running concurrent executable and clean up resources.
        
        Args:
            None
            
        Returns:
            None: Stops the executable and cleans up state
            
        Raises:
            ValueError: If executable is not currently running
        """
        try:
            if self.is_current_running:
                self.running_future.cancel()
                self.is_current_running = False
                logger.debug(f"ConcurrentExecutable '{self.name}' stopped successfully")
            else:
                raise ValueError("Concurrent executable is not running")
        except Exception:
            self.is_current_running = False
            self.running_future = None
        finally:
            self.running_future = None
            self.is_current_running = False



from abc import ABC, abstractmethod
from typing import Dict, Tuple, Any, List, TYPE_CHECKING, Iterator, Optional
import re
import time
import concurrent.futures

if TYPE_CHECKING:
    from .build import Build
    from .concurrent_executable import ConcurrentExecutable


class Blueprint(ABC):
    """Base interface that all user-defined blueprints must implement."""

    @abstractmethod
    def fill(self, parameters: Dict[str, Any]) -> None:
        """
        Fill in the blueprint with the provided parameters.

        Args:
            parameters (Dict[str, Any]): A dictionary of parameters to use to fill in the blueprint
            
        Returns:
            None: Fills the blueprint with provided parameters
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def name(self) -> str:
        """
        Unique identifier for this blueprint.
        
        Args:
            None
            
        Returns:
            str: The unique name of this blueprint
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def usage_context(self) -> str:
        """
        Description of when this blueprint should be used.
        
        Args:
            None
            
        Returns:
            str: Context description for when to use this blueprint
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def purpose(self) -> str:
        """
        Description of what this blueprint is used for.
        
        Args:
            None
            
        Returns:
            str: Purpose description of what this blueprint accomplishes
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def parameter_instructions(self) -> Dict[str, Tuple[str, str]]:
        """
        Map of parameter names to data type and description tuples.
        
        Args:
            None
            
        Returns:
            Dict[str, Tuple[str, str]]: Map of parameter names to (DataType, description) tuples
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def parameter_to_value_map(self) -> Dict[str, Any]:
        """
        Map of parameter names to their filled values.
        
        Args:
            None
            
        Returns:
            Dict[str, Any]: Map of parameter names to their current values
        """
        raise NotImplementedError


class LazyBlueprint:
    """
    A wrapper for blueprints that enables lazy evaluation of parameters.
    
    Instead of immediately resolving $ref references and filling the blueprint,
    this class stores the raw parameters and resolves them on-demand when
    the user calls resolve().
    """
    
    def __init__(
        self,
        blueprint: Blueprint,
        raw_parameters: Dict[str, Any],
        stage_outputs: Dict[str, Dict[str, Any]],
        build_instance: "Build",
        instance_key: str
    ):
        """
        Initialize a lazy blueprint.
        
        Args:
            blueprint (Blueprint): The actual blueprint instance to fill
            raw_parameters (Dict[str, Any]): Parameters with potentially unresolved $ref references
            stage_outputs (Dict[str, Dict[str, Any]]): Dictionary of stage outputs containing ConcurrentExecutables
            build_instance (Build): Build instance to access _resolve_references method
            instance_key (str): The top-level plan key for this blueprint instance
            
        Returns:
            None: Initializes the lazy blueprint instance
        """
        self.blueprint = blueprint
        self.raw_parameters = raw_parameters
        self.stage_outputs = stage_outputs
        self.build_instance = build_instance
        self._instance_key = instance_key
        self._filled = False
        self._dependencies = self._extract_dependencies()
    
    def _extract_dependencies(self) -> List["ConcurrentExecutable"]:
        """
        Extract all ConcurrentExecutable dependencies from raw parameters.
        
        Args:
            None
            
        Returns:
            List[ConcurrentExecutable]: List of concurrent executables that this blueprint depends on
        """
        dependencies = []
        ref_pattern = r'\$ref\.(stage_\d+)\.([a-zA-Z0-9_]+(?:\[-?\d+\])*(?:\.[a-zA-Z0-9_]+(?:\[-?\d+\])*)*)'
        
        def find_refs_in_data(data):
            """
            Recursively find $ref references in nested data structures.
            
            Args:
                data: Data structure to search for references
                
            Returns:
                None: Appends found dependencies to the dependencies list
            """
            if isinstance(data, str):
                matches = re.findall(ref_pattern, data)
                for stage_name, tool_name_with_param in matches:
                    tool_name = tool_name_with_param.split('.')[0]
                    if (stage_name in self.stage_outputs and 
                        tool_name in self.stage_outputs[stage_name]):
                        stage_result = self.stage_outputs[stage_name][tool_name]
                        if hasattr(stage_result, 'fetch_results') and hasattr(stage_result, 'is_current_running'):
                            if stage_result not in dependencies:
                                dependencies.append(stage_result)
            elif isinstance(data, dict):
                for value in data.values():
                    find_refs_in_data(value)
            elif isinstance(data, list):
                for item in data:
                    find_refs_in_data(item)
        
        find_refs_in_data(self.raw_parameters)
        return dependencies
    
    def is_ready(self) -> bool:
        """
        Check if all dependencies are ready (all ConcurrentExecutables are complete).
        
        Args:
            None
            
        Returns:
            bool: True if blueprint can be resolved immediately, False otherwise
        """
        if self._filled:
            return True
            
        for dependency in self._dependencies:
            if not dependency.is_ready():
                return False
        
        return True
    
    def resolve(self) -> None:
        """
        Resolve all $ref references and fill the underlying blueprint.
        
        Args:
            None
            
        Returns:
            None: Fills the underlying blueprint with resolved parameters
            
        Raises:
            ValueError: If blueprint is already filled or resolution fails
        """
        if self._filled:
            raise ValueError(f"Blueprint '{self.blueprint.name}' is already filled")
        
        resolved_parameters = self.build_instance._resolve_references(
            self.raw_parameters, 
            self.stage_outputs
        )
        self.blueprint.fill(resolved_parameters)
        self._filled = True
    
    def is_filled(self) -> bool:
        """
        Check if the blueprint has been resolved and filled.
        
        Args:
            None
            
        Returns:
            bool: True if blueprint has been filled, False otherwise
        """
        return self._filled
    
    @property
    def name(self) -> str:
        """
        Get the name of the underlying blueprint.
        
        Args:
            None
            
        Returns:
            str: Name of the underlying blueprint
        """
        return self.blueprint.name

    @property
    def plan_key(self) -> str:
        """
        Return the unique top-level key used in the build plan for this instance.
        
        Args:
            None
        
        Returns:
            str: The plan key for this blueprint instance (e.g., 'result_blueprint_1').
        """
        return self._instance_key


def blueprints_as_completed(lazy_blueprints: List[LazyBlueprint], max_wait_seconds: Optional[float] = None) -> Iterator[LazyBlueprint]:
    """Yield `LazyBlueprint` instances as dependencies complete, with optional timeout.

    Args:
        lazy_blueprints (List[LazyBlueprint]): Blueprints to monitor for readiness.
        max_wait_seconds (Optional[float]): Maximum seconds to wait before raising `TimeoutError`.

    Returns:
        Iterator[LazyBlueprint]: Iterator of ready blueprints in completion order.
    """
    remaining_blueprints: List[LazyBlueprint] = []
    for lazy_blueprint in lazy_blueprints:
        if lazy_blueprint.is_ready():
            yield lazy_blueprint
        else:
            remaining_blueprints.append(lazy_blueprint)
    if not remaining_blueprints:
        return
    dependency_futures = set()
    for lazy_blueprint in remaining_blueprints:
        for dependency in lazy_blueprint._dependencies:
            if hasattr(dependency, 'running_future') and dependency.running_future is not None and hasattr(dependency.running_future, 'done'):
                dependency_futures.add(dependency.running_future)
    if not dependency_futures:
        return
    yielded_blueprints = set()
    start_time_seconds: float = time.perf_counter()
    while dependency_futures:
        remaining_timeout: Optional[float] = None
        if max_wait_seconds is not None:
            elapsed_seconds: float = time.perf_counter() - start_time_seconds
            remaining_timeout = max(0.0, max_wait_seconds - elapsed_seconds)
            if remaining_timeout == 0.0:
                raise TimeoutError("Blueprint gathering timed out")
        done, pending = concurrent.futures.wait(dependency_futures, timeout=remaining_timeout, return_when=concurrent.futures.FIRST_COMPLETED)
        if not done:
            raise TimeoutError("Blueprint gathering timed out")
        dependency_futures = pending
        for lazy_blueprint in remaining_blueprints:
            if lazy_blueprint not in yielded_blueprints and lazy_blueprint.is_ready():
                yielded_blueprints.add(lazy_blueprint)
                yield lazy_blueprint



from typing import Dict, List, Optional
import logging

from .blueprint import Blueprint
from .logging_utils import log_structured

logger = logging.getLogger(__name__)


class BlueprintRackError(Exception):
    """
    Base exception for BlueprintRack-related errors.
    
    Args:
        message (str): Error message
        **context: Additional context information for the error
        
    Returns:
        None: Exception class for blueprint rack errors
    """

    def __init__(self, message: str, **context):
        """
        Initialize the blueprint rack error with message and context.
        
        Args:
            message (str): Error message
            **context: Additional context information
            
        Returns:
            None: Initializes the exception instance
        """
        super().__init__(message)
        self.context = context


class BlueprintValidationError(BlueprintRackError):
    """Raised when blueprint validation fails."""

    def __init__(
        self,
        issue: str,
        blueprint_type: Optional[str] = None,
        name: Optional[str] = None,
        **context,
    ):
        """
        Initialize the blueprint validation error.
        
        Args:
            issue (str): Description of the validation issue
            blueprint_type (Optional[str]): Type of blueprint that failed validation
            name (Optional[str]): Name of blueprint that failed validation
            **context: Additional context information
            
        Returns:
            None: Initializes the validation error instance
        """
        super().__init__(
            f"Blueprint validation failed: {issue}",
            issue=issue,
            blueprint_type=blueprint_type,
            name=name,
            **context,
        )


class DuplicateBlueprintError(BlueprintRackError):
    """Raised when attempting to add a blueprint with a name that already exists."""

    def __init__(self, name: str, **context):
        """
        Initialize the duplicate blueprint error.
        
        Args:
            name (str): Name of the duplicate blueprint
            **context: Additional context information
            
        Returns:
            None: Initializes the duplicate blueprint error instance
        """
        super().__init__(
            f"Blueprint with name '{name}' already exists", name=name, **context
        )


class BlueprintRack:
    def __init__(self, blueprints: Optional[List[Blueprint]] = None):
        """
        Initialize the BlueprintRack as a container for blueprints available to the Architect.

        Args:
            blueprints (Optional[List[Blueprint]]): A list of Blueprint objects to add to the BlueprintRack
            
        Returns:
            None: Initializes the blueprint rack instance
        """
        self.blueprints: Dict[str, Blueprint] = {}
        if blueprints:
            log_structured(
                logger,
                "info",
                "Initializing BlueprintRack",
                blueprint_count=len(blueprints),
            )
            self._validate_and_add_blueprints(blueprints)
        else:
            log_structured(logger, "info", "Initialized empty BlueprintRack")

    def add_blueprint(self, blueprint: Blueprint) -> None:
        """
        Adds a Blueprint object to the BlueprintRack.

        Args:
            blueprint (Blueprint): The Blueprint object to add to the BlueprintRack
            
        Returns:
            None: Adds the blueprint to the rack
        """
        if not hasattr(blueprint, "name"):
            log_structured(
                logger,
                "error",
                "Failed to add blueprint - missing name attribute",
                blueprint_type=type(blueprint).__name__,
            )
            raise BlueprintValidationError(
                "Blueprint must have a name attribute",
                blueprint_type=type(blueprint).__name__,
            )
        if blueprint.name in self.blueprints:
            log_structured(
                logger,
                "error",
                "Failed to add blueprint - name already exists",
                name=blueprint.name,
            )
            raise DuplicateBlueprintError(blueprint.name)
        self.blueprints[blueprint.name] = blueprint
        log_structured(
            logger, "debug", "Added blueprint to BlueprintRack", name=blueprint.name
        )

    def set_blueprint_list(self, blueprints: List[Blueprint]) -> None:
        """
        Sets the list of Blueprint objects to the BlueprintRack, removing any existing blueprints.

        Args:
            blueprints (List[Blueprint]): A list of Blueprint objects to add to the BlueprintRack
            
        Returns:
            None: Replaces all existing blueprints with the provided list
        """
        old_count = len(self.blueprints)
        self.blueprints = {}
        self._validate_and_add_blueprints(blueprints)
        log_structured(
            logger,
            "info",
            "Updated blueprint list in BlueprintRack",
            old_count=old_count,
            new_count=len(blueprints),
        )

    def find_by_name(self, name: str) -> Optional[Blueprint]:
        """
        Finds a Blueprint object by its name.

        Args:
            name (str): The name of the Blueprint object to find

        Returns:
            Optional[Blueprint]: The Blueprint object if found, otherwise None
        """
        blueprint = self.blueprints.get(name)
        if blueprint:
            log_structured(logger, "debug", "Found blueprint by name", name=name)
            return blueprint
        attempted_name = name
        parsed_name = None
        if '_' in name and name.rsplit('_', 1)[-1].isdigit():
            parsed_name = name.rsplit('_', 1)[0]
            fallback = self.blueprints.get(parsed_name)
            if fallback:
                log_structured(
                    logger,
                    "info",
                    f"Blueprint not found by exact name '{attempted_name}'; using parsed base name '{parsed_name}'",
                    attempted_name=attempted_name,
                    resolved_name=parsed_name,
                    available_blueprints=list(self.blueprints.keys()),
                )
                return fallback
        log_structured(
            logger,
            "warning",
            f"Blueprint not found by name '{attempted_name}' (parsed base '{parsed_name}')",
            name=attempted_name,
            parsed_name=parsed_name,
            available_blueprints=list(self.blueprints.keys()),
        )
        return None

    def remove_blueprint(self, name: str) -> None:
        """
        Removes a Blueprint object from the BlueprintRack.

        Args:
            name (str): The name of the Blueprint object to remove
            
        Returns:
            None: Removes the blueprint from the rack
        """
        if name in self.blueprints:
            del self.blueprints[name]
            log_structured(
                logger, "debug", "Removed blueprint from BlueprintRack", name=name
            )
        else:
            log_structured(
                logger,
                "warning",
                "Attempted to remove non-existent blueprint",
                name=name,
            )

    def open_blueprint(self, name: str) -> Optional[str]:
        """
        Creates a string representation of a Blueprint object formatted for consumption by a Large Language Model.

        Args:
            name (str): The name of the Blueprint object to open

        Returns:
            Optional[str]: A string representation of the Blueprint object, or None if not found
        """
        blueprint: Optional[Blueprint] = self.blueprints.get(name)
        if not blueprint:
            return None
        blueprint_details: str = f"BLUEPRINT NAME: {blueprint.name}\n"
        blueprint_details += f"Usage Context: {blueprint.usage_context}\n"
        blueprint_details += f"Purpose: {blueprint.purpose}\n"
        blueprint_details += "Parameter Instructions:\n"
        for param_name, (
            data_type,
            description,
        ) in blueprint.parameter_instructions.items():
            blueprint_details += f"  {param_name} ({data_type}): {description}\n"
        return blueprint_details

    def open_all_blueprints(self) -> str:
        """
        Opens all Blueprint objects and appends the details to a single string.

        Args:
            None
            
        Returns:
            str: A string representation of all Blueprint objects
        """
        all_blueprint_details: str = "BLUEPRINTS:\n--------------------------------\n"
        for name in self.blueprints.keys():
            blueprint_details: Optional[str] = self.open_blueprint(name)
            if blueprint_details:
                all_blueprint_details += (
                    blueprint_details + "\n--------------------------------\n"
                )
        return all_blueprint_details

    def _validate_and_add_blueprints(self, blueprints: List[Blueprint]) -> None:
        """
        Validates and adds a list of Blueprint objects to the BlueprintRack.

        Args:
            blueprints (List[Blueprint]): A list of Blueprint objects to add to the BlueprintRack
            
        Returns:
            None: Validates and adds blueprints to the rack
            
        Raises:
            BlueprintValidationError: If blueprint validation fails
            DuplicateBlueprintError: If blueprint name already exists
        """
        for blueprint in blueprints:
            if not hasattr(blueprint, "name"):
                log_structured(
                    logger,
                    "error",
                    "Blueprint validation failed - missing name attribute",
                    blueprint_type=type(blueprint).__name__,
                )
                raise BlueprintValidationError(
                    "Blueprint must have a name attribute",
                    blueprint_type=type(blueprint).__name__,
                )
            if blueprint.name in self.blueprints:
                log_structured(
                    logger,
                    "error",
                    "Blueprint validation failed - duplicate name",
                    name=blueprint.name,
                )
                raise DuplicateBlueprintError(blueprint.name)
            self.blueprints[blueprint.name] = blueprint
        log_structured(
            logger,
            "info",
            "Successfully validated and added blueprints",
            count=len(blueprints),
        )



"""Client wrapper for API calls."""

import asyncio
import logging
from typing import List, Dict, Any, Optional, Set, Callable
from concurrent.futures import Future
from openai import AsyncOpenAI, OpenAI
from enum import Enum
from openai.types.chat import ChatCompletion
from concurrent.futures import wait, FIRST_COMPLETED
import importlib
try:
    json = importlib.import_module("ujson")
except Exception:
    import json as json

from .logging_utils import log_structured as log_with_structure

logger = logging.getLogger(__name__)


class ClientType(Enum):
    OPENAI = "openai"
    ANTHROPIC = "anthropic"


class Client:
    """Wrapper for an LLM API client with build plan generation methods."""

    def __init__(
        self, client_type: ClientType, client_api_key: str, use_async: bool, model_name: str, asyncio_event_loop: Optional[asyncio.AbstractEventLoop] = None, token_limit: Optional[int] = None, prompt_cache_key: Optional[str] = None
    ):
        """
        Initialize the client with API credentials and configuration.
        
        Args:
            client_type (ClientType): Type of client to use (OpenAI, Anthropic)
            client_api_key (str): API key for the client
            use_async (bool): Whether to use asynchronous operations
            model_name (str): Name of the model to use
            asyncio_event_loop (Optional[asyncio.AbstractEventLoop]): Event loop for async operations
            token_limit (Optional[int]): Maximum number of tokens to use
            prompt_cache_key (Optional[str]): Cache key for prompt caching optimization
            
        Returns:
            None: Initializes the client instance
        """
        if use_async and asyncio_event_loop is None:
            raise ValueError("Asyncio event loop is required for asyncio execution")
        self.client_type = client_type
        self.client_api_key = client_api_key
        self.model_name = model_name
        self.use_async = use_async
        self.asyncio_event_loop = asyncio_event_loop
        self.token_limit = token_limit
        self.prompt_cache_key = prompt_cache_key or "build_plan_generation"
        self.connected_client = self.initialize_connected_client(use_async)
        log_with_structure(logger, "debug", "Client initialized", client_type=client_type.value, use_async=use_async, model_name=model_name, token_limit=token_limit, prompt_cache_key=self.prompt_cache_key)

    def initialize_connected_client(self, use_async: bool) -> Any:
        """
        Initialize the appropriate client based on client type and async preference.
        
        Args:
            use_async (bool): Whether to create an async or sync client
            
        Returns:
            Any: AsyncOpenAI or OpenAI client instance based on configuration
        """
        if self.client_type == ClientType.OPENAI:
            return (
                AsyncOpenAI(api_key=self.client_api_key)
                if use_async
                else OpenAI(api_key=self.client_api_key)
            )
        else:
            raise ValueError(f"Unsupported client type: {self.client_type}")

    def create_messages_for_build_plan_generation(
        self,
        developer_prompt_base: str,
        developer_prompt_additional_context: Optional[str] = None,
        developer_prompt_previous_attempt_error_message: Optional[str] = None,
        user_prompt: str = "",
    ) -> List[Dict[str, str]]:
        """
        Create structured messages array for build plan generation API call.
        
        Args:
            developer_prompt_base (str): Base system prompt for the developer role
            developer_prompt_additional_context (Optional[str]): Additional context if provided
            developer_prompt_previous_attempt_error_message (Optional[str]): Error from previous attempt
            user_prompt (str): User's request prompt
            
        Returns:
            List[Dict[str, str]]: Formatted messages array for API call
        """
        log_with_structure(logger, "debug", "Creating messages for build plan generation", user_prompt_length=len(user_prompt))
        if self.client_type == ClientType.OPENAI:
            messages = [{"role": "developer", "content": developer_prompt_base}]
            if developer_prompt_additional_context:
                messages.append(
                    {
                        "role": "developer",
                        "content": developer_prompt_additional_context,
                    }
                )
            if developer_prompt_previous_attempt_error_message:
                messages.append(
                    {
                        "role": "developer",
                        "content": developer_prompt_previous_attempt_error_message,
                    }
                )
            if user_prompt:
                messages.append({"role": "user", "content": user_prompt})
            else:
                raise ValueError("User prompt is required")
        else:
            raise ValueError(f"Unsupported client type: {self.client_type}")
        log_with_structure(logger, "debug", "Messages created for build plan generation", message_count=len(messages))
        return messages

    def send_request_to_connected_openai_client(self, messages: List[Dict[str, str]], asyncio_event_loop: Optional[asyncio.AbstractEventLoop] = None) -> str:
        """
        Send request to OpenAI client using sync or async mode based on configuration.
        
        Args:
            messages (List[Dict[str, str]]): Formatted messages for API call
            asyncio_event_loop (Optional[asyncio.AbstractEventLoop]): Event loop for async operations
            
        Returns:
            str: Response content from OpenAI API
        """
        log_with_structure(logger, "debug", "Sending request to OpenAI", use_async=self.use_async, model=self.model_name)
        if self.use_async:
            if asyncio_event_loop is None:
                raise ValueError("Asyncio event loop is required for asyncio execution")
            log_with_structure(logger, "debug", "Creating async OpenAI request")
            api_params = {
                "model": self.model_name,
                "messages": messages,
                "temperature": 0.0,
                "response_format": { "type": "json_object" },
                "service_tier": "priority",
                "timeout": 60.0
            }
            if self.token_limit is not None:
                api_params["max_completion_tokens"] = self.token_limit
            if self.prompt_cache_key:
                api_params["prompt_cache_key"] = self.prompt_cache_key
            try:
                async_response: Future[ChatCompletion] = asyncio.run_coroutine_threadsafe(
                    self.connected_client.chat.completions.create(**api_params),
                    asyncio_event_loop
                )
            except TypeError as cache_error:
                if "prompt_cache_key" in str(cache_error) and self.prompt_cache_key:
                    log_with_structure(logger, "warning", "prompt_cache_key not supported, retrying without", error=str(cache_error))
                    api_params.pop("prompt_cache_key", None)
                    async_response: Future[ChatCompletion] = asyncio.run_coroutine_threadsafe(
                        self.connected_client.chat.completions.create(**api_params),
                        asyncio_event_loop
                    )
                else:
                    raise
            log_with_structure(logger, "debug", "Waiting for async OpenAI response")
            result = async_response.result().choices[0].message.content or ""
            log_with_structure(logger, "debug", "Received async OpenAI response", response_length=len(result))
            return result
        else:
            log_with_structure(logger, "debug", "Creating sync OpenAI request")
            api_params = {
                "model": self.model_name,
                "messages": messages,
                "temperature": 0.0,
                "response_format": { "type": "json_object" },
                "service_tier": "priority",
                "timeout": 60.0
            }
            if self.token_limit is not None:
                api_params["max_completion_tokens"] = self.token_limit
            sync_response: ChatCompletion = self.connected_client.chat.completions.create(**api_params)
            result = sync_response.choices[0].message.content or ""
            log_with_structure(logger, "debug", "Received sync OpenAI response", response_length=len(result))
            return result

    def send_request_to_connected_client(self, messages: List[Dict[str, str]], asyncio_event_loop: Optional[asyncio.AbstractEventLoop] = None) -> str:
        """
        Send request to the connected client based on client type.
        
        Args:
            messages (List[Dict[str, str]]): Formatted messages for API call
            asyncio_event_loop (Optional[asyncio.AbstractEventLoop]): Event loop for async operations
            
        Returns:
            str: Response content from the API
        """
        if self.client_type == ClientType.OPENAI:
            return self.send_request_to_connected_openai_client(messages, asyncio_event_loop)
        else:
            raise ValueError(f"Unsupported client type: {self.client_type}")

    async def send_request_to_connected_client_async(self, messages: List[Dict[str, str]]) -> str:
        """
        Send request to connected client using native async operations.
        
        Args:
            messages (List[Dict[str, str]]): Formatted messages for API call
            
        Returns:
            str: Response content from the API
        """
        log_with_structure(logger, "debug", "Sending async request to connected client")
        if self.client_type == ClientType.OPENAI:
            api_params = {
                "model": self.model_name,
                "messages": messages,
                "temperature": 0.0,
                "response_format": { "type": "json_object" },
                "service_tier": "priority",
                "timeout": 60.0
            }
            if self.token_limit is not None:
                api_params["max_completion_tokens"] = self.token_limit
            response: ChatCompletion = await self.connected_client.chat.completions.create(**api_params)
            result = response.choices[0].message.content or ""
            log_with_structure(logger, "debug", "Received async client response", response_length=len(result))
            return result
        else:
            raise ValueError(f"Unsupported client type: {self.client_type}")

    def send_build_plan_generation_request(
        self,
        messages: List[Dict[str, str]],
        number_of_concurrent_requests: int = 1,
    ) -> str:
        """
        Send build plan generation request with optional concurrent execution.
        
        Args:
            messages (List[Dict[str, str]]): Formatted messages for API call
            number_of_concurrent_requests (int): Number of concurrent requests to make
            
        Returns:
            str: First successful response from concurrent requests
        """
        log_with_structure(logger, "debug", "Starting build plan generation request", concurrent_requests=number_of_concurrent_requests)
        if number_of_concurrent_requests == 1:
            log_with_structure(logger, "debug", "Using single request mode")
            result = self.send_request_to_connected_client(messages, self.asyncio_event_loop)
            log_with_structure(logger, "debug", "Build plan generation request completed", result_length=len(result))
            return result
        log_with_structure(logger, "debug", "Using concurrent request mode")
        if self.asyncio_event_loop is None:
            raise ValueError("Asyncio event loop is required when number_of_concurrent_requests > 1")
        futures = []
        for request_index in range(number_of_concurrent_requests):
            log_with_structure(logger, "debug", "Creating concurrent request", request_index=request_index)
            future: Future[str] = asyncio.run_coroutine_threadsafe(
                self.send_request_to_connected_client_async(messages),
                self.asyncio_event_loop
            )
            futures.append(future)
        log_with_structure(logger, "debug", "Waiting for first concurrent request to complete")
        done, pending = wait(futures, return_when=FIRST_COMPLETED)
        log_with_structure(logger, "debug", "First concurrent request completed, cancelling others", pending_count=len(pending))
        for future in pending:
            future.cancel()
        completed_future = list(done)[0]
        result = completed_future.result()
        log_with_structure(logger, "debug", "Concurrent build plan generation completed", result_length=len(result))
        return result

    def send_build_plan_generation_request_streaming(
        self,
        messages: List[Dict[str, str]],
        stage_callback: Callable[[str, Dict[str, Any]], None],
        blueprint_callback: Optional[Callable[[str, Dict[str, Any]], None]] = None,
        number_of_concurrent_requests: int = 1,
    ) -> str:
        """
        Send streaming build plan generation request with stage-by-stage execution callbacks.
        
        Args:
            messages (List[Dict[str, str]]): Formatted messages for API call
            stage_callback (Callable[[str, Dict[str, Any]], None]): Called when complete stages are detected in stream
            blueprint_callback (Optional[Callable[[str, Dict[str, Any]], None]]): Called when complete blueprints are detected in stream
            number_of_concurrent_requests (int): Number of concurrent requests to make
            
        Returns:
            str: Complete response content from API
        """
        log_with_structure(logger, "debug", "Starting streaming build plan generation request", concurrent_requests=number_of_concurrent_requests)
        if number_of_concurrent_requests > 1:
            raise ValueError("Streaming mode does not support concurrent requests")
        if self.client_type != ClientType.OPENAI:
            raise ValueError("Streaming only supported for OpenAI client type")
        log_with_structure(logger, "debug", "Using streaming request mode")
        api_params = {
            "model": self.model_name,
            "messages": messages,
            "temperature": 0.0,
            "response_format": {"type": "json_object"},
            "stream": True,
            "service_tier": "priority",
            "timeout": 60.0
        }
        if self.token_limit is not None:
            api_params["max_completion_tokens"] = self.token_limit
        if self.use_async:
            if self.asyncio_event_loop is None:
                raise ValueError("Asyncio event loop is required for asyncio execution")
            return self._stream_async_request(api_params, stage_callback, blueprint_callback)
        else:
            return self._stream_sync_request(api_params, stage_callback, blueprint_callback)

    def _stream_sync_request(self, api_params: Dict[str, Any], stage_callback: Callable[[str, Dict[str, Any]], None], blueprint_callback: Optional[Callable[[str, Dict[str, Any]], None]]) -> str:
        """
        Handle streaming sync request with incremental JSON parsing.
        
        Args:
            api_params (Dict[str, Any]): API parameters for the request
            stage_callback (Callable[[str, Dict[str, Any]], None]): Callback for complete stages
            blueprint_callback (Optional[Callable[[str, Dict[str, Any]], None]]): Callback for complete blueprints
            
        Returns:
            str: Complete response content
        """
        full_content = ""
        parsed_stages: Set[str] = set()
        parsed_blueprints: Set[str] = set()
        stream = self.connected_client.chat.completions.create(**api_params)
        for chunk in stream:
            if chunk.choices[0].delta.content:
                content_piece = chunk.choices[0].delta.content
                full_content += content_piece
                if "}" in content_piece:
                    self._check_for_complete_objects(full_content, parsed_stages, stage_callback, parsed_blueprints, blueprint_callback)
        return full_content

    def _stream_async_request(self, api_params: Dict[str, Any], stage_callback: Callable[[str, Dict[str, Any]], None], blueprint_callback: Optional[Callable[[str, Dict[str, Any]], None]]) -> str:
        """
        Handle streaming async request with incremental JSON parsing.
        
        Args:
            api_params (Dict[str, Any]): API parameters for the request
            stage_callback (Callable[[str, Dict[str, Any]], None]): Callback for complete stages
            blueprint_callback (Optional[Callable[[str, Dict[str, Any]], None]]): Callback for complete blueprints
            
        Returns:
            str: Complete response content
        """
        async def stream_coroutine():
            full_content = ""
            parsed_stages: Set[str] = set()
            parsed_blueprints: Set[str] = set()
            stream = await self.connected_client.chat.completions.create(**api_params)
            async for chunk in stream:
                if chunk.choices[0].delta.content:
                    content_piece = chunk.choices[0].delta.content
                    full_content += content_piece
                    if "}" in content_piece:
                        self._check_for_complete_objects(full_content, parsed_stages, stage_callback, parsed_blueprints, blueprint_callback)
            return full_content
        if self.asyncio_event_loop is None:
            raise ValueError("Asyncio event loop is required for asyncio execution")
        future = asyncio.run_coroutine_threadsafe(stream_coroutine(), self.asyncio_event_loop)
        return future.result()

    def _check_for_complete_objects(self, content: str, parsed_stages: Set[str], stage_callback: Callable[[str, Dict[str, Any]], None], parsed_blueprints: Set[str], blueprint_callback: Optional[Callable[[str, Dict[str, Any]], None]]) -> None:
        """
        Check content for complete top-level objects (stages and blueprints) and trigger callbacks.
        
        Args:
            content (str): Current accumulated content
            parsed_stages (Set[str]): Set of already parsed stage names
            stage_callback (Callable[[str, Dict[str, Any]], None]): Callback to invoke for new complete stages
            parsed_blueprints (Set[str]): Set of already parsed blueprint names
            blueprint_callback (Optional[Callable[[str, Dict[str, Any]], None]]): Callback for new complete blueprints
            
        Returns:
            None: Calls callbacks for newly detected complete objects
        """
        found_objects = list(self._iter_top_level_objects(content))
        for key_name, obj_text in found_objects:
            if key_name.startswith("stage_"):
                if key_name in parsed_stages:
                    continue
                try:
                    wrapped = "{" + f'"{key_name}":' + obj_text + "}"
                    parsed = json.loads(wrapped)
                    parsed_stages.add(key_name)
                    stage_callback(key_name, parsed[key_name])
                    log_with_structure(logger, "debug", "Streaming stage detected, starting execution", stage_name=key_name)
                except Exception:
                    continue
            else:
                if blueprint_callback is None or key_name in parsed_blueprints:
                    continue
                try:
                    wrapped = "{" + f'"{key_name}":' + obj_text + "}"
                    parsed = json.loads(wrapped)
                    parsed_blueprints.add(key_name)
                    blueprint_callback(key_name, parsed[key_name])
                    log_with_structure(logger, "debug", "Streaming blueprint detected, processing", blueprint_name=key_name)
                except Exception:
                    continue

    def _iter_top_level_objects(self, content: str):
        """
        Iterate complete top-level JSON key/object pairs from a partial or complete JSON document.
        
        Args:
            content (str): Accumulated JSON text
        
        Yields:
            Tuple[str, str]: Key name and raw JSON object text for that key
        """
        content = content.strip()
        if not content or not content.startswith('{'):
            return
        position = 1
        content_length = len(content)
        while position < content_length:
            position = self._skip_whitespace(content, position)
            if position >= content_length or content[position] == '}':
                break
            key, position = self._parse_json_key(content, position)
            if key is None:
                break
            position = self._skip_to_value(content, position)
            if position >= content_length:
                break
            if content[position] != '{':
                position = self._skip_non_object_value(content, position)
                continue
            obj_value, position = self._parse_object_value(content, position)
            if obj_value is not None:
                yield key, obj_value

    def _skip_whitespace(self, content: str, position: int) -> int:
        """
        Skip whitespace characters from current position.
        
        Args:
            content (str): JSON content string
            position (int): Current position in content
            
        Returns:
            int: New position after skipping whitespace
        """
        while position < len(content) and content[position].isspace():
            position += 1
        return position

    def _parse_json_key(self, content: str, position: int) -> tuple:
        """
        Parse JSON key from current position.
        
        Args:
            content (str): JSON content string
            position (int): Current position in content
            
        Returns:
            tuple: (key_name, new_position) or (None, position) if parsing fails
        """
        if content[position] != '"':
            return None, position
        position += 1
        key_start = position
        while position < len(content) and content[position] != '"':
            if content[position] == '\\':
                position += 2
            else:
                position += 1
        if position >= len(content):
            return None, position
        key = content[key_start:position]
        position += 1
        return key, position

    def _skip_to_value(self, content: str, position: int) -> int:
        """
        Skip whitespace and colon to reach the value part.
        
        Args:
            content (str): JSON content string
            position (int): Current position in content
            
        Returns:
            int: Position at start of value or end of content
        """
        position = self._skip_whitespace(content, position)
        if position >= len(content) or content[position] != ':':
            return len(content)
        position += 1
        return self._skip_whitespace(content, position)

    def _skip_non_object_value(self, content: str, position: int) -> int:
        """
        Skip over non-object JSON values.
        
        Args:
            content (str): JSON content string
            position (int): Current position in content
            
        Returns:
            int: Position after the skipped value
        """
        nesting_depth = 0
        inside_string = False
        while position < len(content):
            current_char = content[position]
            if inside_string:
                if current_char == '"' and (position == 0 or content[position-1] != '\\'):
                    inside_string = False
            else:
                if current_char == '"':
                    inside_string = True
                elif current_char in ['{', '[']:
                    nesting_depth += 1
                elif current_char in ['}', ']']:
                    nesting_depth -= 1
                elif current_char == ',' and nesting_depth == 0:
                    position += 1
                    break
                elif current_char == '}' and nesting_depth == -1:
                    break
            position += 1
        return position

    def _parse_object_value(self, content: str, position: int) -> tuple:
        """
        Parse complete JSON object value from current position.
        
        Args:
            content (str): JSON content string
            position (int): Current position in content
            
        Returns:
            tuple: (object_text, new_position) or (None, position) if incomplete
        """
        obj_start = position
        brace_nesting = 0
        inside_string = False
        while position < len(content):
            current_char = content[position]
            if inside_string:
                if current_char == '"' and (position == 0 or content[position-1] != '\\'):
                    inside_string = False
            else:
                if current_char == '"':
                    inside_string = True
                elif current_char == '{':
                    brace_nesting += 1
                elif current_char == '}':
                    brace_nesting -= 1
                    if brace_nesting == 0:
                        obj_end = position + 1
                        obj_value = content[obj_start:obj_end]
                        position = self._skip_whitespace(content, obj_end)
                        if position < len(content) and content[position] == ',':
                            position += 1
                        return obj_value, position
            position += 1
        return None, position




"""
Shared logging utilities for the architect_ai package.
"""
import logging
from typing import Optional


def log_structured(
    logger: logging.Logger,
    level: str,
    message: str,
    correlation_id: Optional[str] = None,
    perf_time: Optional[float] = None,
    **kwargs,
) -> None:
    """
    Helper function for structured logging using extra parameters.
    
    Args:
        logger (logging.Logger): The logger instance to use
        level (str): The log level ('debug', 'info', 'warning', 'error', 'critical')
        message (str): The log message
        correlation_id (Optional[str]): Optional correlation ID for request tracing
        perf_time (Optional[float]): Optional performance timing in seconds
        **kwargs: Additional structured data to include in the log
        
    Returns:
        None: Logs the structured message with extra data
    """
    RESERVED_ATTRS = {
        'name', 'msg', 'args', 'levelname', 'levelno', 'pathname', 'filename',
        'module', 'lineno', 'funcName', 'created', 'msecs', 'relativeCreated',
        'thread', 'threadName', 'processName', 'process', 'getMessage',
        'exc_info', 'exc_text', 'stack_info'
    }
    extra_data = {}
    for key, value in kwargs.items():
        if key in RESERVED_ATTRS:
            extra_data[f"data_{key}"] = value
        else:
            extra_data[key] = value
    if correlation_id:
        extra_data["correlation_id"] = correlation_id
    if perf_time is not None:
        extra_data["perf_time"] = round(perf_time, 3)
    getattr(logger, level)(message, extra=extra_data)



"""Tool-related exception classes for architect_ai package."""

from typing import List, Optional


class ToolError(Exception):
    """Base exception for tool-related errors."""
    
    def __init__(self, message: str, tool_name: Optional[str] = None, **context):
        """
        Initialize the tool error with message and context.
        
        Args:
            message (str): Error message
            tool_name (Optional[str]): Name of the tool that caused the error
            **context: Additional context information
            
        Returns:
            None: Initializes the tool error instance
        """
        super().__init__(message)
        self.tool_name = tool_name
        self.context = context


class PrecallableToolNotFoundError(ToolError):
    """Raised when a precallable tool is not found."""
    
    def __init__(self, tool_name: str, available_tools: Optional[List[str]] = None, **context):
        """
        Initialize the precallable tool not found error.
        
        Args:
            tool_name (str): Name of the tool that was not found
            available_tools (Optional[List[str]]): List of available tool names
            **context: Additional context information
            
        Returns:
            None: Initializes the precallable tool not found error instance
        """
        super().__init__(
            f"Precallable tool '{tool_name}' not found",
            tool_name=tool_name,
            available_tools=available_tools or [],
            **context
        )


class PrecallableToolExecutionError(ToolError):
    """Raised when getting result from precallable tool fails."""
    
    def __init__(self, tool_name: str, original_error: Exception, **context):
        """
        Initialize the precallable tool execution error.
        
        Args:
            tool_name (str): Name of the tool that failed to execute
            original_error (Exception): The original exception that caused the failure
            **context: Additional context information
            
        Returns:
            None: Initializes the precallable tool execution error instance
        """
        super().__init__(
            f"Failed to get result from precallable tool '{tool_name}': {original_error}",
            tool_name=tool_name,
            original_error=str(original_error),
            original_error_type=type(original_error).__name__,
            **context
        )


class PrecallableToolRuntimeError(ToolError):
    """Raised when there's a runtime issue with precallable tools (e.g., no event loop)."""
    
    def __init__(self, issue: str, tool_name: Optional[str] = None, **context):
        """
        Initialize the precallable tool runtime error.
        
        Args:
            issue (str): Description of the runtime issue
            tool_name (Optional[str]): Name of the tool with runtime issue
            **context: Additional context information
            
        Returns:
            None: Initializes the precallable tool runtime error instance
        """
        super().__init__(
            f"Precallable tool runtime error: {issue}",
            tool_name=tool_name,
            issue=issue,
            **context
        )



"""
Architect AI - An AI-powered architecture planning system.
"""

from .architect import Architect
from .blueprint import Blueprint
from .blueprint_rack import (
    BlueprintRack,
    # Blueprint rack exceptions
    BlueprintRackError,
    BlueprintValidationError,
    DuplicateBlueprintError,
)
from .tool import Tool
from .concurrent_executable import ConcurrentExecutable
from .execution_mode import ExecutionMode
from .toolbox import ToolBox
from .client import Client
from .build import Build
# Note: concurrency_wrapper module doesn't exist in current system
from .exceptions import (
    # Architect exceptions
    ArchitectError,
    PrecallableToolError,
    BuildPlanParsingError,
    BlueprintNotFoundError,
    ToolExecutionModeError,
    ToolExecutionError,
    ReferenceResolutionError,
    InvalidReferencePathError,
    StageValidationError,
    BuildPlanGenerationError,
    ToolPicklingError,
    # Tool exceptions
    ToolError,
    PrecallableToolNotFoundError,
    PrecallableToolExecutionError,
    PrecallableToolRuntimeError,
    # Toolbox exceptions
    ToolBoxError,
    ToolValidationError,
    DuplicateToolError,
)

__version__ = "0.1.0"
__all__ = [
    # Core classes
    "Architect",
    "Blueprint", 
    "BlueprintRack",
    "Tool",
    "ConcurrentExecutable",
    "ExecutionMode",
    "ToolBox",
    "Client",
    "Build",
    # Architect exceptions
    "ArchitectError",
    "PrecallableToolError",
    "BuildPlanParsingError", 
    "BlueprintNotFoundError",
    "ToolExecutionModeError",
    "ToolExecutionError",
    "ReferenceResolutionError",
    "InvalidReferencePathError",
    "StageValidationError",
    "BuildPlanGenerationError",
    "ToolPicklingError",
    # Blueprint rack exceptions
    "BlueprintRackError",
    "BlueprintValidationError",
    "DuplicateBlueprintError",
    # Tool exceptions
    "ToolError",
    "PrecallableToolNotFoundError",
    "PrecallableToolExecutionError", 
    "PrecallableToolRuntimeError",
    # Toolbox exceptions
    "ToolBoxError",
    "ToolValidationError",
    "DuplicateToolError",
]



from enum import Enum


class ExecutionMode(Enum):  
    """Defines the type of concurrent execution."""

    IMMEDIATE = "immediate"
    ASYNCIO = "asyncio"
    THREAD = "thread"
    PROCESS = "process"


"""Architect-related exception classes for architect_ai package."""

from typing import Optional


class ArchitectError(Exception):
    """Base exception for all Architect-related errors."""
    
    def __init__(self, message: str, correlation_id: Optional[str] = None, **context):
        """
        Initialize the architect error with message, correlation ID and context.
        
        Args:
            message (str): Error message
            correlation_id (Optional[str]): Optional correlation ID for request tracing
            **context: Additional context information
            
        Returns:
            None: Initializes the architect error instance
        """
        super().__init__(message)
        self.correlation_id = correlation_id
        self.context = context
    
    def __str__(self):
        """
        Return string representation of the error with correlation ID if available.
        
        Args:
            None
            
        Returns:
            str: String representation of the error
        """
        base = super().__str__()
        if self.correlation_id:
            base += f" [correlation_id: {self.correlation_id}]"
        return base


class PrecallableToolError(ArchitectError):
    """Raised when precallable tool startup fails."""
    
    def __init__(self, message: str, tool_count: Optional[int] = None, **context):
        """
        Initialize the precallable tool error.
        
        Args:
            message (str): Error message
            tool_count (Optional[int]): Number of tools that failed
            **context: Additional context information
            
        Returns:
            None: Initializes the precallable tool error instance
        """
        super().__init__(
            f"Precallable tool error: {message}",
            tool_count=tool_count,
            **context
        )


class BuildPlanParsingError(ArchitectError):
    """Raised when build plan JSON parsing fails."""
    
    def __init__(self, json_error: str, build_plan_preview: Optional[str] = None, **context):
        """
        Initialize the build plan parsing error.
        
        Args:
            json_error (str): JSON parsing error message
            build_plan_preview (Optional[str]): Preview of the build plan that failed to parse
            **context: Additional context information
            
        Returns:
            None: Initializes the build plan parsing error instance
        """
        super().__init__(
            f"Failed to parse build plan JSON: {json_error}",
            json_error=json_error,
            build_plan_preview=build_plan_preview,
            **context
        )


class BlueprintNotFoundError(ArchitectError):
    """Raised when a requested blueprint is not found in the blueprint rack."""
    
    def __init__(self, blueprint_name: str, **context):
        """
        Initialize the blueprint not found error.
        
        Args:
            blueprint_name (str): Name of the blueprint that was not found
            **context: Additional context information
            
        Returns:
            None: Initializes the blueprint not found error instance
        """
        super().__init__(
            f"Blueprint '{blueprint_name}' not found in blueprint rack",
            blueprint_name=blueprint_name,
            **context
        )


class ToolExecutionModeError(ArchitectError):
    """Raised when there's a mismatch between tool execution mode and implementation."""
    
    def __init__(self, tool_name: str, execution_mode: str, issue: str, **context):
        """
        Initialize the tool execution mode error.
        
        Args:
            tool_name (str): Name of the tool with execution mode issue
            execution_mode (str): The problematic execution mode
            issue (str): Description of the execution mode issue
            **context: Additional context information
            
        Returns:
            None: Initializes the tool execution mode error instance
        """
        super().__init__(
            f"Tool '{tool_name}' execution mode '{execution_mode}' error: {issue}",
            tool_name=tool_name,
            execution_mode=execution_mode,
            issue=issue,
            **context
        )


class ToolExecutionError(ArchitectError):
    """Raised when tool execution fails during stage execution."""
    
    def __init__(self, tool_name: str, stage_name: str, original_error: Exception, **context):
        """
        Initialize the tool execution error.
        
        Args:
            tool_name (str): Name of the tool that failed to execute
            stage_name (str): Name of the stage where execution failed
            original_error (Exception): The original exception that caused the failure
            **context: Additional context information
            
        Returns:
            None: Initializes the tool execution error instance
        """
        super().__init__(
            f"Tool execution failed for '{tool_name}' in stage '{stage_name}': {original_error}",
            tool_name=tool_name,
            stage_name=stage_name,
            original_error=str(original_error),
            original_error_type=type(original_error).__name__,
            **context
        )


class ReferenceResolutionError(ArchitectError):
    """Raised when reference resolution fails."""
    
    def __init__(self, reference_path: str, issue: str, **context):
        """
        Initialize the reference resolution error.
        
        Args:
            reference_path (str): Path of the reference that failed to resolve
            issue (str): Description of the resolution issue
            **context: Additional context information
            
        Returns:
            None: Initializes the reference resolution error instance
        """
        super().__init__(
            f"Failed to resolve reference '{reference_path}': {issue}",
            **context
        )
        self.reference_path = reference_path
        self.issue = issue


class InvalidReferencePathError(ReferenceResolutionError):
    """Raised when a reference path has invalid format."""
    
    def __init__(self, reference_path: str, **context):
        """
        Initialize the invalid reference path error.
        
        Args:
            reference_path (str): The invalid reference path
            **context: Additional context information
            
        Returns:
            None: Initializes the invalid reference path error instance
        """
        super().__init__(
            reference_path,
            "Invalid reference path format",
            **context
        )


class StageValidationError(ReferenceResolutionError):
    """Raised when stage reference validation fails (e.g., forward references)."""
    
    def __init__(self, reference_path: str, current_stage: int, referenced_stage: int, **context):
        """
        Initialize the stage validation error.
        
        Args:
            reference_path (str): Path of the reference that failed validation
            current_stage (int): Current stage number
            referenced_stage (int): Stage number being referenced
            **context: Additional context information
            
        Returns:
            None: Initializes the stage validation error instance
        """
        super().__init__(
            reference_path,
            f"Stage {current_stage} cannot reference stage {referenced_stage}. Stages can only reference outputs from previous stages.",
            current_stage=current_stage,
            referenced_stage=referenced_stage,
            **context
        )


class BuildPlanGenerationError(ArchitectError):
    """Raised when build plan generation fails after maximum attempts."""
    
    def __init__(self, max_attempts: int, **context):
        """
        Initialize the build plan generation error.
        
        Args:
            max_attempts (int): Maximum number of attempts that were made
            **context: Additional context information
            
        Returns:
            None: Initializes the build plan generation error instance
        """
        super().__init__(
            f"Failed to generate build plan after {max_attempts} attempts",
            max_attempts=max_attempts,
            **context
        )


class ToolPicklingError(ArchitectError):
    """Raised when a PROCESS-mode tool instance cannot be pickled for execution."""
    
    def __init__(self, tool_name: str, original_error: Optional[Exception] = None, **context):
        """
        Initialize the tool pickling error.
        
        Args:
            tool_name (str): Name of the tool that failed to be pickled
            original_error (Optional[Exception]): The original pickling error
            **context: Additional context information
            
        Returns:
            None: Initializes the tool pickling error instance
        """
        super().__init__(
            f"Tool '{tool_name}' is not pickleable for PROCESS execution: {original_error}",
            tool_name=tool_name,
            original_error=str(original_error) if original_error else None,
            original_error_type=type(original_error).__name__ if original_error else None,
            **context
        )


class ToolNotFoundError(ArchitectError):
    """Raised when a referenced tool is not found in the toolbox during stage execution."""
    
    def __init__(self, tool_name: str, stage_name: str, available_tools: Optional[list] = None, **context):
        """
        Initialize the tool not found error.
        
        Args:
            tool_name (str): Name of the tool that was not found
            stage_name (str): Name of the stage where the tool was needed
            available_tools (Optional[list]): List of available tools
            **context: Additional context information
            
        Returns:
            None: Initializes the tool not found error instance
        """
        super().__init__(
            f"Tool '{tool_name}' not found in toolbox while executing stage '{stage_name}'",
            tool_name=tool_name,
            stage_name=stage_name,
            available_tools=available_tools or [],
            **context
        )


class ExecutableFailedToStartError(ArchitectError):
    """
    Raised when a concurrent executable fails to start.
    
    Args:
        message (str): Error message describing the failure
        
    Returns:
        None: Exception class for concurrent executable start failures
    """
    pass


class BuildGenerationError(ArchitectError):
    """
    Raised when build plan generation fails.
    
    Args:
        message (str): Error message describing the generation failure
        
    Returns:
        None: Exception class for build generation failures
    """
    pass


import logging
import uuid
import asyncio
import time
from .client import Client
from .toolbox import ToolBox
from .blueprint_rack import BlueprintRack
from .concurrent_executable import ConcurrentExecutable
from .build import Build
from .blueprint import LazyBlueprint
from .exceptions.architect_exceptions import ExecutableFailedToStartError
from .logging_utils import log_structured as log_with_structure
from typing import List, Dict, Optional, Tuple
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor


logger = logging.getLogger(__name__)


class Architect:
    """
    An expert systems architect that generates and executes build plans using available tools and blueprints.
    
    Args:
        client (Client): Client for communicating with AI model
        model_name (str): Name of the AI model to use
        toolbox (ToolBox): Collection of available tools
        blueprint_rack (BlueprintRack): Collection of available blueprints
        initial_concurrent_executables (Optional[List[ConcurrentExecutable]]): Initial executables to start
        asyncio_event_loop (Optional[asyncio.AbstractEventLoop]): Event loop for async operations
        thread_pool_executor (Optional[ThreadPoolExecutor]): Thread pool for concurrent execution
        process_pool_executor (Optional[ProcessPoolExecutor]): Process pool for concurrent execution
        
    Returns:
        Architect: Configured architect instance ready to generate responses
    """
    build_plan_generation_base_prompt = """
    You are an expert systems architect, named Arch, with 30 years of experience in the technology industry. Your primary objective is to generate a build plan that will outline exactly how to make use of the various tools at your disposal to meet the product owner's demands (the user with which you will be interacting is the product owner). You must also ensure that this information is documented correctly by filling in the parameters of any blueprints that are applicable to the customer's request. Each tool in your toolbox comes with a guide that details when to use it, what it does, and how to format the input parameters to get your desired output parameters. Each blueprint on the blueprint rack comes with a guide that details when to use it and how to correctly fill in the parameters to ensure that your work has been correctly documented. 

    The final deliverable will be a JSON representation of the build plan constructed such that there is no room for ambiguity. There are two types of "top level" keys that should be present in the JSON. First, there should be a series of numbered "stage" keys ('stage_1', 'stage_2', etc.) that each contain a list of tools and their respective input parameters. Each tool contained within a stage will be executed in parallel. Once all the tools in a stage have returned, the next stage will begin. Subsequent stages will have access to the outputs of the previous stages. Where possible, make full use of parallel execution to maximize efficiency, including multiple tools in the same stage. To include tool calls within a stage, you should use the following format:

    {{
        "stage_1": {{
            "tool_name_1": {{
                "input_param_1": "text_value_1",
                "input_param_2": 8,
            }}
            "tool_name_2": {{
                "input_param_1": "text_value_1",
            }}
        }}
        ...
    }}

    Tool calls in subsequent stages will often need to reference the outputs of the previous stages. To make use of this ability, you should use the format $ref.stage_name.tool_name.output_param_name anywhere within strings, lists, or nested structures. References can be mixed with regular values. See the following example:

    {{
        "stage_1": {{
            "tool_name_1": {{
                "input_param_1": "text_value_1",
            }}
        }}
        "stage_2": {{
            "tool_name_2": {{
                "input_param_1": "Processing result: $ref.stage_1.tool_name_1.output_param_1",
                "input_param_2": ["item1", 42, "$ref.stage_1.tool_name_1.output_param_2"],
                "input_param_3": "$ref.stage_1.tool_name_1.output_param_1"
            }}
        }}
    }}

    The other top level keys in the JSON object must be the names of the blueprint(s) that are applicable to the customer's request. These will be mapped to a list of input parameters for that respective blueprint. Again, each blueprint will come with instructions as to when to use it and how to correctly fill in the parameters. Here is an example of how the final JSON object should look. However, keep in mind that the number of stages and blueprints required will vary based on the customer's request.

    {{
        "stage_1": {{
            "tool_name_1": {{
                "input_param_1": "text_value_1",
            }}
            ...
        }}
        "stage_2": {{
            "tool_name_2": {{
                "input_param_1": "$ref.stage_1.tool_name_1.output_param_1"
            }}
            "tool_name_3": {{
                "input_param_1": "text_value_1",
            }}
            ...
        }}
        ...
        "blueprint_1": {{
            "input_param_1": "text_value_1",
            "input_param_2": "$ref.stage_2.tool_name_2.output_param_1"
        }}
        "blueprint_2": {{
            "input_param_1": "$ref.stage_2.tool_name_3.output_param_3",
            "input_param_2": "$ref.stage_1.tool_name_1.output_param_1"
        }}
        ...
    }}

    HERE IS YOUR TOOLBOX:
    --------------------------------
    {toolbox}
    --------------------------------

    HERE IS YOUR BLUEPRINT RACK:
    --------------------------------
    {blueprint_rack}
    --------------------------------

    CRITICAL INSTRUCTIONS:
    - NEVER use any input or output parameter names for tools or blueprints that are not explicitly listed in the toolbox or the blueprint rack, respectively.
    - NEVER produce an incomplete output or use any placeholders of your own invention. If you are unable to call a tool or fill a blueprint accurately, you still must make sure that your build plan will compile successfully.
    - Input parameters can be simple values, strings with embedded references, lists containing mixed values and references, or nested structures. References using $ref.stage.tool.output format will be resolved recursively.
    - NEVER try to index a variable like $ref.stage_1.tool_1.output_list[0:2] - references cannot be sliced.
    - ALWAYS use only the precise tool names from the toolbox. However, if the you must call the same tool multiple times in thesame stage, append _X to the tool name, where X is the index of the tool call.
    - ALWAYS use only the precise blueprint names from the blueprint rack. However, if the you must call the same blueprint multiple times in thesame stage, append _X to the blueprint name, where X is the index of the blueprint call.
    - Note that for string data type input parameters for tools or blueprints, you can include multiple references to the same or different tools embedded in the string. 
    - Additionally, you are allowed to references nested parameters of tool outputs by using dot notation like: $ref.stage_1.tool_1.output_param_1.nested_param_1.
    - You may also use the following syntax to reference a list index: $ref.stage_1.tool_1.output_param_1[0]. Negative indices are also allowed, but you MAY NOT splice or do any other complex operations.
    - You can do deep nested references like $ref.stage_1.tool_1.output_param_1.nested_param_1.nested_param_list_2[0].nested_param_3 if needed, but keep things simple when possible.
    - DO NOT include any fancy references using operations I have not approved you for, such as splicing.
    - NEVER include blueprints within stages. They can ONLY be top level keys no matter what.
    - NEVER reference the same stage within that stage (for example a tool in stage 2 cannot reference another tool in stage 2). You must separate the call into another stage.
    - NEVER include a blueprint inside of a stage. EVER!

    ADDITIONAL FORMATTING EXAMPLE FOR REFERENCE:
    --------------------------------
    {{
        "stage_1": {{
            "tool_name_1": {{
                "input_param_1": "text_value_1",
            }}
            ...
        }}
        "stage_2": {{
            "tool_name_2": {{
                "input_param_1": "$ref.stage_1.tool_name_1.output_param_1",
                "input_param_2": "$ref.stage_1.tool_name_1.output_param_list[0].nested_param_in_list_1[-1][1]"
            }}
            "tool_name_3": {{
                "input_param_1": "text_value_1",
            }}
            ...
        }}
        ...
        "blueprint_1": {{
            "input_param_1": "text_value_1",
            "input_param_2": "$ref.stage_2.tool_name_2.output_param_1"
        }}
        "blueprint_2": {{
            "input_param_1": "$ref.stage_2.tool_name_3.output_param_3",
            "input_param_2": "$ref.stage_1.tool_name_1.output_param_1"
        }}
        ...
    }}
    --------------------------------

    """

    build_plan_additional_context_prompt_wrapper = """
    URGENT UPDATE: The product owner has provided additional information! All new information should be utilized as needed, and any new instructions must be followed exactly. Do your best to fulfill all requirements, which will generally be possible, but in the event of a contradiction, these updates must take precedence over the original requirements:

    --------------------------------
    {additional_context_prompt}
    --------------------------------

    Regardless of any new information, you must still return a valid build plan that will compile successfully according to the previous formatting instructions.
    """

    build_plan_previous_attempt_error_message_prompt_wrapper = """
    CORRECTION NEEDED: The product owner has informed us that on the previous attempt to generate and executea build plan, the following error message was returned:
    --------------------------------
    {previous_attempt_error_message}
    --------------------------------
    
    This error message should be taken into account when generating the next build plan to avoid making the same mistake. Never include details from this message in the build plan, but use it to guide the structure of the build plan so that the same error doesn't happen again.
    """

    user_prompt_wrapper = """
    Hi, Arch. Product owner here. Generate a build plan for the following customer request:
    --------------------------------
    {customer_request}
    --------------------------------

    Here is the conversation history with the customer:
    --------------------------------
    {conversation_history}
    --------------------------------

    And once again here is the initial request that you should focus on fulfilling:
    --------------------------------
    {customer_request}
    --------------------------------
    """

    def __init__(
        self,
        client: Client,
        model_name: str,
        toolbox: ToolBox,
        blueprint_rack: BlueprintRack,
        initial_concurrent_executables: Optional[List[ConcurrentExecutable]] = None,
        asyncio_event_loop: Optional[asyncio.AbstractEventLoop] = None,
        thread_pool_executor: Optional[ThreadPoolExecutor] = None,
        process_pool_executor: Optional[ProcessPoolExecutor] = None,
        enable_streaming_execution: bool = False,
    ):
        """
        Initialize the Architect with required components and execution backends.
        
        Args:
            client (Client): Client class instance for communicating with an LLM provider
            model_name (str): Name of the AI model to use
            toolbox (ToolBox): Collection of available tools
            blueprint_rack (BlueprintRack): Collection of available blueprints
            initial_concurrent_executables (Optional[List[ConcurrentExecutable]]): Initial executables to start
            asyncio_event_loop (Optional[asyncio.AbstractEventLoop]): Event loop for async operations
            thread_pool_executor (Optional[ThreadPoolExecutor]): Thread pool for concurrent execution
            process_pool_executor (Optional[ProcessPoolExecutor]): Process pool for concurrent execution
            enable_streaming_execution (bool): Enable streaming JSON parsing and execution
            
        Returns:
            None: Initializes the architect instance
        """
        if not any([asyncio_event_loop, thread_pool_executor, process_pool_executor]):
            raise ValueError("At least one execution backend must be provided")
        self.client: Client = client
        self.toolbox: ToolBox = toolbox
        self.blueprint_rack: BlueprintRack = blueprint_rack
        self.concurrent_executables: List[ConcurrentExecutable] = (
            initial_concurrent_executables
            if initial_concurrent_executables is not None
            else []
        )
        self.initial_concurrent_executables_started: List[ConcurrentExecutable] = []
        self.architect_id = str(uuid.uuid4())
        self.asyncio_event_loop: Optional[asyncio.AbstractEventLoop] = (
            asyncio_event_loop
        )
        self.thread_pool_executor: Optional[ThreadPoolExecutor] = thread_pool_executor
        self.process_pool_executor: Optional[ProcessPoolExecutor] = (
            process_pool_executor
        )
        self.enable_streaming_execution: bool = enable_streaming_execution
        log_with_structure(
            logger,
            "debug",
            "Architect initalized",
            architect_id=self.architect_id,
            toolbox_count=len(self.toolbox.tools),
            blueprint_rack_count=len(self.blueprint_rack.blueprints),
            concurrent_executable_count=len(self.concurrent_executables),
        )

    def _start_concurrent_executables(self):
        """
        Start all concurrent executables and track which ones started successfully.
        
        Args:
            None
            
        Returns:
            None: Modifies initial_concurrent_executables_started list
        """
        for executable in self.concurrent_executables:
            try:
                executable.start()
                self.initial_concurrent_executables_started.append(executable)
            except ExecutableFailedToStartError as start_error:
                log_with_structure(
                    logger,
                    "error",
                    "Failed to start concurrent executable",
                    architect_id=self.architect_id,
                    executable_name=executable.name,
                    error=start_error,
                )
        log_with_structure(
            logger,
            "debug",
            "Concurrent executables started",
            architect_id=self.architect_id,
            executable_count=len(self.initial_concurrent_executables_started),
        )

    def _generate_build(
        self,
        customer_request: str,
        conversation_history: str,
        additional_context_prompt: str,
        build_correlation_id: str,
        start_time: float,
        previous_attempt_error_message: Optional[str] = None,
        number_of_concurrent_requests_per_call_to_client: int = 1,
    ) -> Build:
        """
        Generate a build plan using the AI client and available tools and blueprints.
        
        Args:
            customer_request (str): The customer's request to fulfill
            conversation_history (str): Previous conversation context
            additional_context_prompt (str): Additional context to include
            build_correlation_id (str): Unique ID for tracking this build
            start_time (float): Performance counter start time
            previous_attempt_error_message (Optional[str]): Error from previous attempt if retrying
            number_of_concurrent_requests_per_call_to_client (int): Number of concurrent requests to make
            
        Returns:
            Build: Generated build plan ready for execution
        """
        developer_prompt_base: str = self.build_plan_generation_base_prompt.format(
            toolbox=self.toolbox.open_all_tools(),
            blueprint_rack=self.blueprint_rack.open_all_blueprints(),
        )
        developer_prompt_additional_context: str = (
            self.build_plan_additional_context_prompt_wrapper.format(
                additional_context_prompt=additional_context_prompt,
            )
        )
        developer_prompt_previous_attempt_error_message: Optional[str] = (
            self.build_plan_previous_attempt_error_message_prompt_wrapper.format(
                previous_attempt_error_message=previous_attempt_error_message
            )
            if previous_attempt_error_message
            else None
        )
        user_prompt: str = self.user_prompt_wrapper.format(
            conversation_history=conversation_history,
            customer_request=customer_request,
        )
        messages: List[Dict[str, str]] = (
            self.client.create_messages_for_build_plan_generation(
                developer_prompt_base=developer_prompt_base,
                developer_prompt_additional_context=developer_prompt_additional_context,
                developer_prompt_previous_attempt_error_message=developer_prompt_previous_attempt_error_message,
                user_prompt=user_prompt,
            )
        )
        log_with_structure(
            logger,
            "info",
            "Starting build plan generation",
            architect_id=self.architect_id,
            build_correlation_id=build_correlation_id,
            time_since_generate_response_called=time.perf_counter() - start_time,
            previous_attempt_error_message=previous_attempt_error_message,
        )
        build_plan: str = ""
        plan: str = ""
        if self.enable_streaming_execution:
            build = Build(
                "",
                build_correlation_id,
                streaming_mode=True,
                toolbox=self.toolbox,
                blueprint_rack=self.blueprint_rack,
                concurrent_executables=self.initial_concurrent_executables_started,
                correlation_id=build_correlation_id,
                asyncio_event_loop=self.asyncio_event_loop,
                thread_pool_executor=self.thread_pool_executor,
                process_pool_executor=self.process_pool_executor
            )
            plan = self.client.send_build_plan_generation_request_streaming(
                messages,
                build.stream_stage_callback,
                build.stream_blueprint_callback,
                number_of_concurrent_requests_per_call_to_client,
            )
            build.build_plan = plan
            build.finalize_streaming_execution()
        else:
            plan = self.client.send_build_plan_generation_request(
                messages, number_of_concurrent_requests_per_call_to_client
            )
            build = Build(plan, build_correlation_id)
        build_plan = plan
        log_with_structure(
            logger,
            "info",
            "Build plan generated",
            architect_id=self.architect_id,
            build_correlation_id=build_correlation_id,
            time_since_generate_response_called=time.perf_counter() - start_time,
            build_plan=build_plan,
        )
        return build

    def generate_response(
        self,
        customer_request: str,
        conversation_history: str,
        additional_context_prompt: str,
        max_attempts_to_generate_build_plan: int = 1,
        number_of_concurrent_requests_per_call_to_client: int = 1,
    ) -> Tuple[Build, List[ConcurrentExecutable], List[LazyBlueprint]]:
        """
        Generate and execute a response to the customer request with retry logic.
        
        Args:
            customer_request (str): The customer's request to fulfill
            conversation_history (str): Previous conversation context
            additional_context_prompt (str): Additional context to include
            max_attempts_to_generate_build_plan (int): Maximum retry attempts
            number_of_concurrent_requests_per_call_to_client (int): Number of concurrent requests to make
            
        Returns:
            Tuple[Build, List[ConcurrentExecutable], List[LazyBlueprint]]: Build result, started executables, and lazy blueprints
        """
        build_correlation_id: str = str(uuid.uuid4())
        start_time: float = time.perf_counter()
        self._start_concurrent_executables()
        previous_attempt_error_message: Optional[str] = None
        for attempt in range(max_attempts_to_generate_build_plan):
            try:
                build: Build = self._generate_build(
                    customer_request,
                    conversation_history,
                    additional_context_prompt,
                    build_correlation_id,
                    start_time,
                    previous_attempt_error_message,
                    number_of_concurrent_requests_per_call_to_client,
                )
                log_with_structure(
                    logger,
                    "debug",
                    "Build plan ready",
                    architect_id=self.architect_id,
                    build_correlation_id=build_correlation_id,
                    build_plan_preview=build.build_plan[:200]
                )
                if not self.enable_streaming_execution:
                    build_start_time = time.perf_counter()
                    build.run(
                        self.toolbox,
                        self.blueprint_rack,
                        self.initial_concurrent_executables_started,
                        build_correlation_id,
                        self.asyncio_event_loop,
                        self.thread_pool_executor,
                        self.process_pool_executor
                    )
                    build_execution_time = time.perf_counter() - build_start_time
                    log_with_structure(
                        logger,
                        "debug",
                        f"Build execution completed in {build_execution_time} seconds",
                        architect_id=self.architect_id,
                        build_correlation_id=build_correlation_id,
                        build_execution_time_seconds=build_execution_time,
                        time_since_generate_response_called=time.perf_counter() - start_time,
                    )
                lazy_blueprints = build.get_lazy_blueprints() or []
                return build, self.initial_concurrent_executables_started, lazy_blueprints
            except Exception as build_error:
                log_with_structure(
                    logger,
                    "error",
                    "Failed to generate or execute build plan",
                    architect_id=self.architect_id,
                    build_correlation_id=build_correlation_id,
                    time_since_generate_response_called=time.perf_counter()
                    - start_time,
                    max_attempts_to_generate_build_plan=max_attempts_to_generate_build_plan,
                    current_attempt=attempt + 1,
                    exception_type=type(build_error).__name__,
                    exception_message=str(build_error),
                    error=build_error,
                )
                logger.error("Full build execution failure traceback:", exc_info=True)
                if attempt < max_attempts_to_generate_build_plan - 1:
                    log_with_structure(
                        logger,
                        "info",
                        "Retrying build plan generation",
                        architect_id=self.architect_id,
                        build_correlation_id=build_correlation_id,
                        time_since_generate_response_called=time.perf_counter()
                        - start_time,
                        max_attempts_to_generate_build_plan=max_attempts_to_generate_build_plan,
                        current_attempt=attempt + 1,
                        error=build_error,
                    )
                    previous_attempt_error_message = str(build_error)

        log_with_structure(
            logger,
            "error",
            "Failed to generate build plan after max attempts",
            architect_id=self.architect_id,
            build_correlation_id=build_correlation_id,
            time_since_generate_response_called=time.perf_counter() - start_time,
            max_attempts_to_generate_build_plan=max_attempts_to_generate_build_plan,
        )
        failed_build = Build("", build_correlation_id)
        return failed_build, self.initial_concurrent_executables_started, []



"""Build plan representation and execution logic."""

import importlib
try:
    json = importlib.import_module("ujson")
except Exception:
    import json as json
import logging
import asyncio
import re
from typing import Dict, Any, Optional, List
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor

from .logging_utils import log_structured as log_with_structure
from .toolbox import ToolBox
from .blueprint_rack import BlueprintRack
from .concurrent_executable import ConcurrentExecutable
from .blueprint import LazyBlueprint
from .exceptions.architect_exceptions import (
    ToolNotFoundError,
    ReferenceResolutionError,
    InvalidReferencePathError,
    StageValidationError,
)

logger = logging.getLogger(__name__)


class Build:
    """Represents a build plan with execution capabilities."""
    
    def __init__(self, build_plan: str, build_correlation_id: str, streaming_mode: bool = False, **streaming_context):
        """Initialize build with plan and correlation ID.
        
        Args:
            build_plan (str): JSON string containing the build plan
            build_correlation_id (str): Unique identifier for this build instance
            streaming_mode (bool): Enable streaming execution mode
            **streaming_context: Additional context for streaming mode
            
        Returns:
            None: Constructor returns None
        """
        self.build_plan = build_plan
        self.build_correlation_id = build_correlation_id
        self._parsed_plan: Optional[Dict[str, Any]] = None
        self.filled_blueprints: Optional[List[Any]] = None
        self.lazy_blueprints: Optional[List[LazyBlueprint]] = None
        self.stage_outputs: Optional[Dict[str, Dict[str, Any]]] = None
        self.streaming_mode = streaming_mode
        self.streaming_context = streaming_context
        if streaming_mode:
            self.streaming_stage_outputs: Dict[str, Dict[str, Any]] = {}
            self.streaming_stage_futures: Dict[str, Any] = {}
    
    def get_parsed_plan(self) -> Dict[str, Any]:
        """Parse and return the build plan as a dictionary.
        
        Args:
            None
            
        Returns:
            Dict[str, Any]: Parsed build plan dictionary
        """
        if self._parsed_plan is None:
            try:
                self._parsed_plan = json.loads(self.build_plan)
            except Exception as json_decode_error:
                log_with_structure(
                    logger,
                    "error",
                    "Failed to parse build plan JSON",
                    build_correlation_id=self.build_correlation_id,
                    error=str(json_decode_error),
                    build_plan_text=self.build_plan,
                )
                raise BuildPlanParsingError(f"Invalid JSON in build plan: {json_decode_error}") from json_decode_error
        return self._parsed_plan
    
    def get_stages(self) -> Dict[str, Dict[str, Any]]:
        """Extract and return all stage configurations from the build plan.
        
        Args:
            None
            
        Returns:
            Dict[str, Dict[str, Any]]: Dictionary of stage configurations keyed by stage name
        """
        parsed_plan = self.get_parsed_plan()
        stages = {}
        for plan_key, plan_value in parsed_plan.items():
            if plan_key.startswith("stage_"):
                stages[plan_key] = plan_value
        return stages
    
    def get_blueprints(self) -> Dict[str, Dict[str, Any]]:
        """Extract and return all blueprint configurations from the build plan.
        
        Args:
            None
            
        Returns:
            Dict[str, Dict[str, Any]]: Dictionary of blueprint configurations keyed by blueprint name
        """
        parsed_plan = self.get_parsed_plan()
        blueprints = {}
        for plan_key, plan_value in parsed_plan.items():
            if not plan_key.startswith("stage_"):
                blueprints[plan_key] = plan_value
        return blueprints
    
    def get_lazy_blueprints(self) -> Optional[List[LazyBlueprint]]:
        """Return the lazy blueprints created during build execution.
        
        Args:
            None
            
        Returns:
            Optional[List[LazyBlueprint]]: List of lazy blueprints or None if not executed
        """
        return self.lazy_blueprints
    
    def execute(
        self,
        toolbox: ToolBox,
        blueprint_rack: BlueprintRack,
        concurrent_executables: Optional[List[Any]] = None,
        correlation_id: Optional[str] = None,
        asyncio_event_loop: Optional[asyncio.AbstractEventLoop] = None,
        thread_pool_executor: Optional[ThreadPoolExecutor] = None,
        process_pool_executor: Optional[ProcessPoolExecutor] = None,
    ) -> None:
        """Execute the build plan using provided resources and store results in instance variables.
        
        Args:
            toolbox (ToolBox): Collection of available tools for execution
            blueprint_rack (BlueprintRack): Collection of available blueprints
            concurrent_executables (Optional[List[Any]]): Optional list of concurrent executables
            correlation_id (Optional[str]): Optional correlation identifier
            asyncio_event_loop (Optional[asyncio.AbstractEventLoop]): Optional asyncio event loop
            thread_pool_executor (Optional[ThreadPoolExecutor]): Optional thread pool executor
            process_pool_executor (Optional[ProcessPoolExecutor]): Optional process pool executor
            
        Returns:
            None: Results are stored in instance variables filled_blueprints, lazy_blueprints, and stage_outputs
        """
        self._initialize_execution_state()
        self._collect_concurrent_results(concurrent_executables)
        stage_outputs = self._execute_all_stages(
            toolbox, concurrent_executables, correlation_id, 
            asyncio_event_loop, thread_pool_executor, process_pool_executor
        )
        lazy_blueprints = self._create_lazy_blueprints(blueprint_rack, stage_outputs)
        self._finalize_execution(lazy_blueprints, stage_outputs)
    
    def _initialize_execution_state(self) -> None:
        """Initialize execution state by clearing previous results.
        
        Args:
            None
            
        Returns:
            None: Clears filled_blueprints, lazy_blueprints, and stage_outputs instance variables
        """
        self.filled_blueprints = None
        self.lazy_blueprints = None
        self.stage_outputs = None

    def _collect_concurrent_results(self, concurrent_executables: Optional[List[Any]]) -> Dict[str, Any]:
        """Collect results from concurrent executables if available.
        
        Args:
            concurrent_executables (Optional[List[Any]]): Optional list of concurrent executables to collect results from
            
        Returns:
            Dict[str, Any]: Dictionary mapping executable names to their results
        """
        concurrent_results: Dict[str, Any] = {}
        if concurrent_executables:
            for executable in concurrent_executables:
                try:
                    if hasattr(executable, "is_current_running") and executable.is_current_running:
                        concurrent_results[executable.name] = executable.fetch_results()
                except Exception as execution_error:
                    log_with_structure(
                        logger,
                        "error",
                        "Failed to get concurrent executable result",
                        build_correlation_id=self.build_correlation_id,
                        executable_name=getattr(executable, "name", None),
                        exception_type=type(execution_error).__name__,
                        exception_message=str(execution_error),
                        error=execution_error,
                    )
        return concurrent_results

    def _execute_all_stages(
        self,
        toolbox: ToolBox,
        concurrent_executables: Optional[List[Any]],
        correlation_id: Optional[str],
        asyncio_event_loop: Optional[asyncio.AbstractEventLoop],
        thread_pool_executor: Optional[ThreadPoolExecutor],
        process_pool_executor: Optional[ProcessPoolExecutor],
    ) -> Dict[str, Dict[str, Any]]:
        """Execute all stages in sorted order and return their outputs.
        
        Args:
            toolbox (ToolBox): Collection of available tools
            concurrent_executables (Optional[List[Any]]): Optional concurrent executables
            correlation_id (Optional[str]): Optional correlation identifier
            asyncio_event_loop (Optional[asyncio.AbstractEventLoop]): Optional asyncio event loop
            thread_pool_executor (Optional[ThreadPoolExecutor]): Optional thread pool executor
            process_pool_executor (Optional[ProcessPoolExecutor]): Optional process pool executor
            
        Returns:
            Dict[str, Dict[str, Any]]: Dictionary mapping stage names to their tool outputs
        """
        stages = self.get_stages()
        stage_outputs: Dict[str, Dict[str, Any]] = {}
        for stage_name in sorted(stages.keys()):
            stage_config = stages[stage_name]
            stage_results = self._execute_stage(
                stage_name,
                stage_config,
                stage_outputs,
                concurrent_executables,
                toolbox,
                correlation_id,
                asyncio_event_loop,
                thread_pool_executor,
                process_pool_executor,
            )
            stage_outputs[stage_name] = stage_results
        return stage_outputs

    def _create_lazy_blueprints(
        self, blueprint_rack: BlueprintRack, stage_outputs: Dict[str, Dict[str, Any]]
    ) -> List[LazyBlueprint]:
        """Create lazy blueprints from blueprint configurations in the build plan.
        
        Args:
            blueprint_rack (BlueprintRack): Collection of available blueprints
            stage_outputs (Dict[str, Dict[str, Any]]): Results from executed stages
            
        Returns:
            List[LazyBlueprint]: List of created lazy blueprints
        """
        blueprint_configs = self.get_blueprints()
        lazy_blueprints: List[LazyBlueprint] = []
        for blueprint_name, blueprint_params in blueprint_configs.items():
            blueprint = self._find_blueprint_with_fallback(blueprint_rack, blueprint_name)
            if blueprint:
                lazy_blueprint = self._create_single_lazy_blueprint(
                    blueprint, blueprint_name, blueprint_params, stage_outputs
                )
                if lazy_blueprint:
                    lazy_blueprints.append(lazy_blueprint)
            else:
                self._log_blueprint_not_found(blueprint_name)
        return lazy_blueprints

    def _find_blueprint_with_fallback(self, blueprint_rack: BlueprintRack, blueprint_name: str):
        """Find blueprint by name with fallback to parsed name.
        
        Args:
            blueprint_rack (BlueprintRack): Collection of available blueprints
            blueprint_name (str): Name of blueprint to find
            
        Returns:
            Blueprint or None: Found blueprint instance or None if not found
        """
        blueprint = blueprint_rack.find_by_name(blueprint_name)
        if not blueprint:
            parsed_blueprint_name = self._parse_name_with_suffix(blueprint_name)
            if parsed_blueprint_name:
                log_with_structure(
                    logger,
                    "debug",
                    "Attempting to find blueprint with parsed name",
                    build_correlation_id=self.build_correlation_id,
                    original_blueprint_name=blueprint_name,
                    parsed_blueprint_name=parsed_blueprint_name
                )
                blueprint = blueprint_rack.find_by_name(parsed_blueprint_name)
                if blueprint:
                    log_with_structure(
                        logger,
                        "debug",
                        "Successfully found blueprint using parsed name",
                        build_correlation_id=self.build_correlation_id,
                        original_blueprint_name=blueprint_name,
                        parsed_blueprint_name=parsed_blueprint_name
                    )
        return blueprint

    def _create_single_lazy_blueprint(
        self, blueprint, blueprint_name: str, blueprint_params, stage_outputs: Dict[str, Dict[str, Any]]
    ) -> Optional[LazyBlueprint]:
        """Create a single lazy blueprint with error handling.
        
        Args:
            blueprint: Blueprint instance to wrap
            blueprint_name (str): Name of the blueprint
            blueprint_params: Parameters for the blueprint
            stage_outputs (Dict[str, Dict[str, Any]]): Stage execution results
            
        Returns:
            Optional[LazyBlueprint]: Created lazy blueprint or None if creation failed
        """
        try:
            lazy_blueprint = LazyBlueprint(
                blueprint=blueprint,
                raw_parameters=blueprint_params,
                stage_outputs=stage_outputs,
                build_instance=self,
                instance_key=blueprint_name
            )
            log_with_structure(
                logger,
                "debug",
                "Created lazy blueprint",
                build_correlation_id=self.build_correlation_id,
                blueprint_name=blueprint_name,
                dependency_count=len(lazy_blueprint._dependencies),
            )
            return lazy_blueprint
        except Exception as blueprint_error:
            log_with_structure(
                logger,
                "error",
                "Failed to create lazy blueprint",
                build_correlation_id=self.build_correlation_id,
                blueprint_name=blueprint_name,
                error=str(blueprint_error),
            )
            return None

    def _log_blueprint_not_found(self, blueprint_name: str) -> None:
        """Log warning when blueprint is not found.
        
        Args:
            blueprint_name (str): Name of the blueprint that was not found
            
        Returns:
            None: Logs warning message
        """
        log_with_structure(
            logger,
            "warning",
            "Blueprint not found",
            build_correlation_id=self.build_correlation_id,
            blueprint_name=blueprint_name,
        )

    def _finalize_execution(
        self, lazy_blueprints: List[LazyBlueprint], stage_outputs: Dict[str, Dict[str, Any]]
    ) -> None:
        """Finalize execution by setting results and resolving remaining executables.
        
        Args:
            lazy_blueprints (List[LazyBlueprint]): Created lazy blueprints
            stage_outputs (Dict[str, Dict[str, Any]]): Stage execution results
            
        Returns:
            None: Sets instance variables and finalizes stage outputs
        """
        self.lazy_blueprints = lazy_blueprints
        self.filled_blueprints = []
        self.stage_outputs = stage_outputs
        self._finalize_stage_outputs()

    def _finalize_stage_outputs(self):
        """Ensure all ConcurrentExecutables in stage_outputs are resolved to their results.
        
        Args:
            None
            
        Returns:
            None: Resolves ConcurrentExecutables in stage_outputs to their actual results
        """
        if self.stage_outputs is None:
            return
        for stage_name, stage_data in self.stage_outputs.items():
            for tool_name, tool_result in stage_data.items():
                if hasattr(tool_result, 'fetch_results') and hasattr(tool_result, 'is_current_running'):
                    resolved_result = self._fetch_executable_result(tool_result, tool_name, stage_name)
                    self.stage_outputs[stage_name][tool_name] = resolved_result
                    log_with_structure(
                        logger,
                        "debug",
                        "ConcurrentExecutable finalized in stage_outputs",
                        build_correlation_id=self.build_correlation_id,
                        stage_name=stage_name,
                        tool_name=tool_name,
                    )

    def run(
        self,
        toolbox: ToolBox,
        blueprint_rack: BlueprintRack,
        concurrent_executables: Optional[List[Any]] = None,
        correlation_id: Optional[str] = None,
        asyncio_event_loop: Optional[asyncio.AbstractEventLoop] = None,
        thread_pool_executor: Optional[ThreadPoolExecutor] = None,
        process_pool_executor: Optional[ProcessPoolExecutor] = None,
    ) -> None:
        """Execute the build plan - maintained for compatibility with architect.py.
        
        Args:
            toolbox (ToolBox): Collection of available tools for execution
            blueprint_rack (BlueprintRack): Collection of available blueprints
            concurrent_executables (Optional[List[Any]]): Optional list of concurrent executables
            correlation_id (Optional[str]): Optional correlation identifier
            asyncio_event_loop (Optional[asyncio.AbstractEventLoop]): Optional asyncio event loop
            thread_pool_executor (Optional[ThreadPoolExecutor]): Optional thread pool executor
            process_pool_executor (Optional[ProcessPoolExecutor]): Optional process pool executor
            
        Returns:
            None: Executes build plan and stores results in instance variables
        """
        self.filled_blueprints = None
        self.lazy_blueprints = None
        self.stage_outputs = None
        self.execute(
            toolbox, 
            blueprint_rack, 
            concurrent_executables, 
            correlation_id,
            asyncio_event_loop,
            thread_pool_executor,
            process_pool_executor
        )

    def _execute_stage(
        self,
        stage_name: str,
        stage_config: Dict[str, Any],
        stage_outputs: Dict[str, Dict[str, Any]],
        concurrent_executables: Optional[List[Any]],
        toolbox: ToolBox,
        correlation_id: Optional[str],
        asyncio_event_loop: Optional[asyncio.AbstractEventLoop] = None,
        thread_pool_executor: Optional[ThreadPoolExecutor] = None,
        process_pool_executor: Optional[ProcessPoolExecutor] = None,
    ) -> Dict[str, Any]:
        """Execute all tools in a stage in parallel using ConcurrentExecutables.
        
        Args:
            stage_name (str): Name of the stage to execute
            stage_config (Dict[str, Any]): Configuration for the stage
            stage_outputs (Dict[str, Dict[str, Any]]): Outputs from previously executed stages
            concurrent_executables (Optional[List[Any]]): Optional concurrent executables
            toolbox (ToolBox): Collection of available tools
            correlation_id (Optional[str]): Optional correlation identifier
            asyncio_event_loop (Optional[asyncio.AbstractEventLoop]): Optional asyncio event loop
            thread_pool_executor (Optional[ThreadPoolExecutor]): Optional thread pool executor
            process_pool_executor (Optional[ProcessPoolExecutor]): Optional process pool executor
            
        Returns:
            Dict[str, Any]: Dictionary mapping tool names to their ConcurrentExecutable instances
        """
        stage_executables, tool_names = self._create_stage_executables(
            stage_name, stage_config, stage_outputs, concurrent_executables,
            toolbox, asyncio_event_loop, thread_pool_executor, process_pool_executor
        )
        self._start_all_executables(stage_name, stage_executables)
        return self._prepare_stage_results(stage_name, stage_executables, tool_names)
    
    def _create_stage_executables(
        self,
        stage_name: str,
        stage_config: Dict[str, Any],
        stage_outputs: Dict[str, Dict[str, Any]],
        concurrent_executables: Optional[List[Any]],
        toolbox: ToolBox,
        asyncio_event_loop: Optional[asyncio.AbstractEventLoop],
        thread_pool_executor: Optional[ThreadPoolExecutor],
        process_pool_executor: Optional[ProcessPoolExecutor],
    ):
        """Create ConcurrentExecutables for all tools in a stage.
        
        Args:
            stage_name (str): Name of the stage
            stage_config (Dict[str, Any]): Stage configuration
            stage_outputs (Dict[str, Dict[str, Any]]): Previous stage outputs
            concurrent_executables (Optional[List[Any]]): Optional concurrent executables
            toolbox (ToolBox): Available tools
            asyncio_event_loop (Optional[asyncio.AbstractEventLoop]): Optional event loop
            thread_pool_executor (Optional[ThreadPoolExecutor]): Optional thread executor
            process_pool_executor (Optional[ProcessPoolExecutor]): Optional process executor
            
        Returns:
            Tuple: (stage_executables, tool_names) - lists of executables and their names
        """
        stage_executables: List[ConcurrentExecutable] = []
        tool_names: List[str] = []
        for tool_name, tool_params in stage_config.items():
            tool = self._find_tool_with_fallback(toolbox, tool_name, stage_name)
            resolved_params = self._resolve_references(tool_params, stage_outputs)
            executable = self._create_single_executable(
                tool, tool_name, resolved_params, concurrent_executables,
                asyncio_event_loop, thread_pool_executor, process_pool_executor, stage_name
            )
            stage_executables.append(executable)
            tool_names.append(tool_name)
        return stage_executables, tool_names

    def _find_tool_with_fallback(self, toolbox: ToolBox, tool_name: str, stage_name: str):
        """Find tool by name with fallback to parsed name.
        
        Args:
            toolbox (ToolBox): Collection of available tools
            tool_name (str): Name of tool to find
            stage_name (str): Name of current stage (for error reporting)
            
        Returns:
            Tool: Found tool instance
        """
        tool = toolbox.find_by_name(tool_name)
        if not tool:
            parsed_tool_name = self._parse_name_with_suffix(tool_name)
            if parsed_tool_name:
                log_with_structure(
                    logger,
                    "debug", 
                    "Attempting to find tool with parsed name",
                    build_correlation_id=self.build_correlation_id,
                    original_tool_name=tool_name,
                    parsed_tool_name=parsed_tool_name,
                    stage_name=stage_name
                )
                tool = toolbox.find_by_name(parsed_tool_name)
                if tool:
                    log_with_structure(
                        logger,
                        "debug",
                        "Successfully found tool using parsed name", 
                        build_correlation_id=self.build_correlation_id,
                        original_tool_name=tool_name,
                        parsed_tool_name=parsed_tool_name,
                        stage_name=stage_name
                    )
            if not tool:
                raise ToolNotFoundError(tool_name, stage_name, available_tools=list(toolbox.tools.keys()))
        return tool

    def _create_single_executable(
        self, tool, tool_name: str, resolved_params, concurrent_executables: Optional[List[Any]],
        asyncio_event_loop: Optional[asyncio.AbstractEventLoop],
        thread_pool_executor: Optional[ThreadPoolExecutor],
        process_pool_executor: Optional[ProcessPoolExecutor],
        stage_name: str
    ) -> ConcurrentExecutable:
        """Create a single ConcurrentExecutable from a tool with error handling.
        
        Args:
            tool: Tool instance to wrap
            tool_name (str): Name of the tool
            resolved_params: Resolved parameters for the tool
            concurrent_executables (Optional[List[Any]]): Optional concurrent executables
            asyncio_event_loop (Optional[asyncio.AbstractEventLoop]): Optional event loop
            thread_pool_executor (Optional[ThreadPoolExecutor]): Optional thread executor
            process_pool_executor (Optional[ProcessPoolExecutor]): Optional process executor
            stage_name (str): Name of current stage
            
        Returns:
            ConcurrentExecutable: Created executable instance
        """
        try:
            executable = ConcurrentExecutable.from_tool(
                tool=tool,
                parameters=resolved_params,
                initial_concurrent_executables=concurrent_executables,
                asyncio_event_loop=asyncio_event_loop,
                thread_pool_executor=thread_pool_executor,
                process_pool_executor=process_pool_executor,
            )
            log_with_structure(
                logger,
                "debug",
                "Created ConcurrentExecutable for tool",
                build_correlation_id=self.build_correlation_id,
                stage_name=stage_name,
                tool_name=tool_name,
                execution_mode=tool.execution_mode.value,
            )
            return executable
        except Exception as creation_error:
            log_with_structure(
                logger,
                "error",
                "Failed to create ConcurrentExecutable",
                build_correlation_id=self.build_correlation_id,
                stage_name=stage_name,
                tool_name=tool_name,
                error=str(creation_error),
            )
            raise

    def _start_all_executables(self, stage_name: str, stage_executables: List[ConcurrentExecutable]) -> None:
        """Start all ConcurrentExecutables in the stage.
        
        Args:
            stage_name (str): Name of the stage
            stage_executables (List[ConcurrentExecutable]): List of executables to start
            
        Returns:
            None: Starts all executables and logs results
        """
        for executable in stage_executables:
            try:
                executable.start()
                log_with_structure(
                    logger,
                    "debug",
                    "Started ConcurrentExecutable",
                    build_correlation_id=self.build_correlation_id,
                    stage_name=stage_name,
                    tool_name=executable.name,
                )
            except Exception as start_error:
                log_with_structure(
                    logger,
                    "error",
                    "Failed to start ConcurrentExecutable",
                    build_correlation_id=self.build_correlation_id,
                    stage_name=stage_name,
                    tool_name=executable.name,
                    error=str(start_error),
                )
    def _prepare_stage_results(
        self, stage_name: str, stage_executables: List[ConcurrentExecutable], tool_names: List[str]
    ) -> Dict[str, Any]:
        """Prepare stage results from running ConcurrentExecutables.
        
        Args:
            stage_name (str): Name of the stage
            stage_executables (List[ConcurrentExecutable]): List of started executables
            tool_names (List[str]): List of tool names corresponding to executables
            
        Returns:
            Dict[str, Any]: Dictionary mapping tool names to their executable instances
        """
        results: Dict[str, Any] = {}
        for executable, tool_name in zip(stage_executables, tool_names):
            results[tool_name] = executable
            log_with_structure(
                logger,
                "debug",
                "ConcurrentExecutable started and ready for lazy evaluation",
                build_correlation_id=self.build_correlation_id,
                stage_name=stage_name,
                tool_name=tool_name,
            )
        return results
    def _fetch_executable_result(self, executable: Any, tool_name: str, stage_name: str) -> Any:
        """Safely fetch results from a ConcurrentExecutable with error handling.
        
        Args:
            executable (Any): ConcurrentExecutable instance to fetch results from
            tool_name (str): Name of the tool for logging
            stage_name (str): Name of the stage for logging
            
        Returns:
            Any: Result from executable or error dictionary if failed
        """
        try:
            executable_result = executable.fetch_results()
            log_with_structure(
                logger,
                "debug",
                "ConcurrentExecutable result fetched on-demand",
                build_correlation_id=self.build_correlation_id,
                stage_name=stage_name,
                tool_name=tool_name,
            )
            return executable_result
        except Exception as fetch_error:
            log_with_structure(
                logger,
                "error",
                "ConcurrentExecutable failed when fetching results on-demand",
                build_correlation_id=self.build_correlation_id,
                stage_name=stage_name,
                tool_name=tool_name,
                exception_type=type(fetch_error).__name__,
                exception_message=str(fetch_error),
                error=fetch_error,
            )
            logger.error(f"Full ConcurrentExecutable failure traceback for {tool_name}:", exc_info=True)
            return {"error": str(fetch_error)}
    
    def _resolve_references(
        self,
        data: Any,
        stage_outputs: Dict[str, Dict[str, Any]],
        current_stage_num: Optional[int] = None,
    ) -> Any:
        """Resolve $ref references in data structures using stage_outputs, with optional stage validation.
        
        Args:
            data (Any): Data structure potentially containing $ref references
            stage_outputs (Dict[str, Dict[str, Any]]): Outputs from executed stages
            current_stage_num (Optional[int]): Current stage number for validation
            
        Returns:
            Any: Data structure with resolved references
        """
        if isinstance(data, str):
            return self._resolve_string_references(data, stage_outputs, current_stage_num)
        if isinstance(data, dict):
            return {dict_key: self._resolve_references(dict_value, stage_outputs, current_stage_num) for dict_key, dict_value in data.items()}
        if isinstance(data, list):
            return [self._resolve_references(list_item, stage_outputs, current_stage_num) for list_item in data]
        return data
    
    def _resolve_string_references(
        self, data: str, stage_outputs: Dict[str, Dict[str, Any]], current_stage_num: Optional[int]
    ) -> Any:
        """Handle reference resolution for string data types.
        
        Args:
            data (str): String potentially containing references
            stage_outputs (Dict[str, Dict[str, Any]]): Stage execution results
            current_stage_num (Optional[int]): Current stage number for validation
            
        Returns:
            Any: Resolved value (may be string or other type)
        """
        if data.startswith("$ref.") and " " not in data:
            return self._resolve_single_reference(data, stage_outputs, current_stage_num)
        if "$ref." in data:
            return self._resolve_embedded_references(data, stage_outputs, current_stage_num)
        return data
    def _resolve_single_reference(
        self, reference: str, stage_outputs: Dict[str, Dict[str, Any]], current_stage_num: Optional[int]
    ) -> Any:
        """Resolve a single $ref reference.
        
        Args:
            reference (str): Reference string to resolve
            stage_outputs (Dict[str, Dict[str, Any]]): Stage execution results
            current_stage_num (Optional[int]): Current stage number for validation
            
        Returns:
            Any: Resolved value from the referenced location
        """
        reference_parts = reference.split(".")
        if len(reference_parts) < 4:
            raise InvalidReferencePathError(f"Invalid reference path format: {reference}")
        stage_key = reference_parts[1]
        if not stage_key.startswith("stage_"):
            raise InvalidReferencePathError(f"Invalid stage token in reference: {reference}")
        try:
            reference_stage_number = int(stage_key.split("_")[1])
        except Exception:
            raise InvalidReferencePathError(f"Invalid stage number in reference: {reference}")
        if current_stage_num is not None:
            if reference_stage_number > current_stage_num:
                raise StageValidationError(reference, current_stage_num, reference_stage_number)
            if reference_stage_number == current_stage_num:
                raise StageValidationError(reference, current_stage_num, current_stage_num)
        tool_name = reference_parts[2]
        output_param = ".".join(reference_parts[3:])
        if stage_key not in stage_outputs or tool_name not in stage_outputs[stage_key]:
            raise ReferenceResolutionError(reference, "missing stage or tool")
        stage_result = stage_outputs[stage_key][tool_name]
        if hasattr(stage_result, 'fetch_results') and hasattr(stage_result, 'is_current_running'):
            # If we're in async context and the executable is ASYNCIO or a future, await it; else block here
            try:
                current_loop = asyncio.get_running_loop()
            except RuntimeError:
                current_loop = None
            if current_loop is not None and getattr(stage_result, 'execution_mode', None) == ExecutionMode.ASYNCIO:
                # We are in event loop; use async fetch via temporary task run
                async def _await_exec():
                    return await stage_result.fetch_results_async()
                stage_result = current_loop.run_until_complete(_await_exec())
            else:
                stage_result = self._fetch_executable_result(stage_result, tool_name, stage_key)
            stage_outputs[stage_key][tool_name] = stage_result
        return self._navigate_output_parameter(stage_result, output_param, reference)
    def _navigate_output_parameter(self, stage_result: Any, output_param: str, reference: str) -> Any:
        """Navigate nested object access with support for list indexing.
        
        Args:
            stage_result (Any): Result object to navigate
            output_param (str): Parameter path to navigate
            reference (str): Original reference for error reporting
            
        Returns:
            Any: Value at the specified parameter path
        """
        current_result = stage_result
        param_parts = output_param.split(".")
        for param_index, param_part in enumerate(param_parts):
            if '[' in param_part and param_part.endswith(']'):
                current_result = self._handle_bracket_notation(
                    current_result, param_part, param_parts, param_index, reference
                )
            else:
                if not isinstance(current_result, dict) or param_part not in current_result:
                    partial_path = ".".join(param_parts[:param_index+1])
                    raise ReferenceResolutionError(reference, f"output parameter '{partial_path}' not found")
                current_result = current_result[param_part]
        return current_result
    def _handle_bracket_notation(
        self, current_result: Any, param_part: str, param_parts: List[str], param_index: int, reference: str
    ) -> Any:
        """Handle bracket notation for list indexing in parameter navigation.
        
        Args:
            current_result (Any): Current navigation result
            param_part (str): Parameter part with bracket notation
            param_parts (List[str]): Full list of parameter parts
            param_index (int): Index of current parameter part
            reference (str): Original reference for error reporting
            
        Returns:
            Any: Result after applying bracket notation indexing
        """
        current_part = param_part
        first_bracket = current_part.find('[')
        base_name = current_part[:first_bracket] if first_bracket > 0 else ""
        if base_name:
            if not isinstance(current_result, dict) or base_name not in current_result:
                partial_path = ".".join(param_parts[:param_index]) + ("." if param_index > 0 else "") + base_name
                raise ReferenceResolutionError(reference, f"output parameter '{partial_path}' not found")
            current_result = current_result[base_name]
        bracket_part = current_part[first_bracket:] if first_bracket >= 0 else current_part
        while '[' in bracket_part and bracket_part.endswith(']'):
            start_index = bracket_part.find('[')
            end_index = bracket_part.find(']', start_index)
            if start_index == -1 or end_index == -1:
                break
            index_content = bracket_part[start_index+1:end_index]
            try:
                list_index = int(index_content)
                if not isinstance(current_result, (list, tuple)):
                    partial_path = ".".join(param_parts[:param_index+1])
                    raise ReferenceResolutionError(reference, f"output parameter '{partial_path}' is not subscriptable")
                current_result = current_result[list_index]
            except ValueError:
                partial_path = ".".join(param_parts[:param_index+1])
                raise ReferenceResolutionError(reference, f"invalid index '{index_content}' in '{partial_path}'")
            except IndexError:
                partial_path = ".".join(param_parts[:param_index+1])
                raise ReferenceResolutionError(reference, f"list index out of range in '{partial_path}'")
            bracket_part = bracket_part[end_index+1:]
            if not bracket_part:
                break
        return current_result
    def _resolve_embedded_references(
        self, data: str, stage_outputs: Dict[str, Dict[str, Any]], current_stage_num: Optional[int]
    ) -> str:
        """Handle embedded references within text strings.
        
        Args:
            data (str): String with embedded references
            stage_outputs (Dict[str, Dict[str, Any]]): Stage execution results
            current_stage_num (Optional[int]): Current stage number for validation
            
        Returns:
            str: String with all references resolved to their values
        """
        log_with_structure(
            logger,
            "debug",
            "Processing string with embedded references",
            build_correlation_id=self.build_correlation_id,
            original_text=data
        )
        reference_pattern = r'\$ref\.(stage_\d+)\.([a-zA-Z0-9_]+(?:\[-?\d+\])*(?:\.[a-zA-Z0-9_]+(?:\[-?\d+\])*)*)\.([a-zA-Z0-9_]+(?:\[-?\d+\])*(?:\.[a-zA-Z0-9_]+(?:\[-?\d+\])*)*)'
        def replace_reference_match(pattern_match):
            full_reference = f"$ref.{pattern_match.group(1)}.{pattern_match.group(2)}.{pattern_match.group(3)}"
            try:
                resolved_value = self._resolve_single_reference(full_reference, stage_outputs, current_stage_num)
                log_with_structure(
                    logger,
                    "debug",
                    "Successfully resolved embedded reference",
                    build_correlation_id=self.build_correlation_id,
                    original_text=data,
                    reference=full_reference,
                    resolved_value=str(resolved_value)
                )
                return str(resolved_value)
            except Exception as resolution_error:
                log_with_structure(
                    logger,
                    "error", 
                    "Failed to resolve embedded reference",
                    build_correlation_id=self.build_correlation_id,
                    original_text=data,
                    reference=full_reference,
                    error=str(resolution_error)
                )
                raise resolution_error
        try:
            result = re.sub(reference_pattern, replace_reference_match, data)
            log_with_structure(
                logger,
                "debug", 
                "Completed embedded reference processing",
                build_correlation_id=self.build_correlation_id,
                original_text=data,
                final_result=result
            )
            return result
        except Exception as processing_error:
            raise processing_error
    def _parse_name_with_suffix(self, name: str) -> Optional[str]:
        """Parse tool/blueprint name with _X suffix and return base name.
        
        Args:
            name (str): Tool or blueprint name potentially with numeric suffix
            
        Returns:
            Optional[str]: Base name without suffix or None if no valid suffix found
        """
        if '_' not in name:
            return None
        name_parts = name.rsplit('_', 1)
        if len(name_parts) != 2:
            return None
        base_name, suffix = name_parts
        if not suffix.isdigit():
            return None
        return base_name
    
    def __repr__(self) -> str:
        """Return string representation of Build instance.
        
        Args:
            None
            
        Returns:
            str: String representation showing correlation ID and plan length
        """
        return f"Build(correlation_id={self.build_correlation_id}, plan_length={len(self.build_plan)})"

    def stream_stage_callback(self, stage_name: str, stage_config: Dict[str, Any]) -> None:
        """
        Callback invoked when a complete stage is detected in streaming mode.
        
        Args:
            stage_name (str): Name of the detected stage
            stage_config (Dict[str, Any]): Configuration for the stage
            
        Returns:
            None: Starts execution of the detected stage
        """
        if not self.streaming_mode:
            return
        log_with_structure(
            logger,
            "info",
            f"Streaming stage detected: {stage_name}",
            build_correlation_id=self.build_correlation_id,
            stage_name=stage_name,
        )
        try:
            toolbox = self.streaming_context.get("toolbox")
            concurrent_executables = self.streaming_context.get("concurrent_executables")
            correlation_id = self.streaming_context.get("correlation_id")
            asyncio_event_loop = self.streaming_context.get("asyncio_event_loop")
            thread_pool_executor = self.streaming_context.get("thread_pool_executor")
            process_pool_executor = self.streaming_context.get("process_pool_executor")
            if toolbox:
                if not hasattr(self, "_streaming_stage_queue"):
                    self._streaming_stage_queue: List[tuple] = []
                # Defer execution to finalize to avoid blocking event loop; preserve order
                self._streaming_stage_queue.append((stage_name, stage_config))
        except Exception as stage_error:
            log_with_structure(
                logger,
                "error",
                "Failed to execute streaming stage",
                build_correlation_id=self.build_correlation_id,
                stage_name=stage_name,
                error=str(stage_error),
                error_type=type(stage_error).__name__,
            )

    # Removed _run_streaming_stage helper and prior-stage waiting; we execute stages immediately on detection

    def stream_blueprint_callback(self, blueprint_name: str, blueprint_params: Dict[str, Any]) -> None:
        """
        Callback invoked when a complete blueprint is detected in streaming mode.
        
        Args:
            blueprint_name (str): Name of the detected blueprint
            blueprint_params (Dict[str, Any]): Parameters for the blueprint
            
        Returns:
            None: Creates a lazy blueprint incrementally for the detected object
        """
        if not self.streaming_mode:
            return
        try:
            if not hasattr(self, "_streaming_lazy_blueprints"):
                self._streaming_lazy_blueprints: List[LazyBlueprint] = []
            blueprint_rack = self.streaming_context.get("blueprint_rack")
            if blueprint_rack:
                # Check without logging warnings
                blueprint = blueprint_rack.blueprints.get(blueprint_name)
                if not blueprint:
                    parsed_name = self._parse_name_with_suffix(blueprint_name)
                    blueprint = blueprint_rack.blueprints.get(parsed_name) if parsed_name else None
                if blueprint:
                    log_with_structure(
                        logger,
                        "info",
                        f"Streaming blueprint detected: {blueprint_name}",
                        build_correlation_id=self.build_correlation_id,
                        blueprint_name=blueprint_name,
                    )
                    lazy_blueprint = LazyBlueprint(
                        blueprint=blueprint,
                        raw_parameters=blueprint_params,
                        stage_outputs=self.streaming_stage_outputs,
                        build_instance=self,
                        instance_key=blueprint_name
                    )
                    self._streaming_lazy_blueprints.append(lazy_blueprint)
                else:
                    log_with_structure(
                        logger,
                        "debug",
                        "Ignoring non-blueprint top-level object during streaming",
                        build_correlation_id=self.build_correlation_id,
                        blueprint_name=blueprint_name,
                    )
        except Exception as blueprint_error:
            log_with_structure(
                logger,
                "error",
                "Failed to process streaming blueprint",
                build_correlation_id=self.build_correlation_id,
                blueprint_name=blueprint_name,
                error=str(blueprint_error),
            )

    def finalize_streaming_execution(self) -> None:
        """
        Finalize streaming execution by creating blueprints and setting final results.
        
        Args:
            None
            
        Returns:
            None: Sets stage_outputs and creates lazy blueprints
        """
        if not self.streaming_mode:
            return
        # Execute any deferred streaming stages now, in order, on the current thread
        if hasattr(self, "_streaming_stage_queue") and self._streaming_stage_queue:
            toolbox: ToolBox = self.streaming_context.get("toolbox")
            concurrent_executables = self.streaming_context.get("concurrent_executables")
            correlation_id = self.streaming_context.get("correlation_id")
            asyncio_event_loop = self.streaming_context.get("asyncio_event_loop")
            thread_pool_executor = self.streaming_context.get("thread_pool_executor")
            process_pool_executor = self.streaming_context.get("process_pool_executor")
            for stage_name, stage_config in self._streaming_stage_queue:
                stage_results = self._execute_stage(
                    stage_name,
                    stage_config,
                    self.streaming_stage_outputs,
                    concurrent_executables,
                    toolbox,
                    correlation_id,
                    asyncio_event_loop,
                    thread_pool_executor,
                    process_pool_executor,
                )
                self.streaming_stage_outputs[stage_name] = stage_results
            self._streaming_stage_queue.clear()
        self.stage_outputs = self.streaming_stage_outputs
        blueprint_rack = self.streaming_context.get("blueprint_rack")
        if blueprint_rack:
            if hasattr(self, "_streaming_lazy_blueprints") and self._streaming_lazy_blueprints:
                self._finalize_execution(self._streaming_lazy_blueprints, self.stage_outputs)
            else:
                # In streaming mode, avoid parsing the full plan at finalize time.
                # If no streamed blueprints were detected, finalize with an empty list.
                self._finalize_execution([], self.stage_outputs)


class BuildPlanParsingError(Exception):
    """Raised when build plan JSON cannot be parsed.
    
    Args:
        Exception: Standard exception base class
        
    Returns:
        None: Exception class for build plan parsing errors
    """
    pass



#!/usr/bin/env python3
"""
DecisionBlueprint - A built-in blueprint for handling conditional logic across multiple architect calls.

This blueprint is used when the architect needs to make dynamic decisions based on tool outputs,
but cannot complete the full conditional flow in a single pass. It stores intermediate results
and descriptions that can be used by the caller to make subsequent architect calls.
"""

from typing import Dict, Tuple, Any, List
from .blueprint import Blueprint


class DecisionBlueprint(Blueprint):
    """
    A blueprint that stores intermediate results for conditional logic scenarios.
    
    This blueprint is designed for situations where the architect can perform some work
    but needs dynamic decision-making based on results. It stores intermediate results
    and descriptions that enable the caller to chain multiple architect calls together.
    
    Example usage:
    - User asks: "Multiply result of 4+5 by 2 if <8, by 4 if >8"  
    - First call: Architect adds 4+5=9, stores in DecisionBlueprint: ["The result of 4 + 5 was 9"]
    - Caller sees result, makes second call with context about the 9 and which path to take
    """
    
    def __init__(self):
        """
        Initialize the decision blueprint.
        
        Args:
            None
            
        Returns:
            None: Initializes the decision blueprint instance
        """
        self._intermediate_results: List[str] = []
        self._filled = False
    
    @property
    def name(self) -> str:
        """
        Unique identifier for this blueprint.
        
        Args:
            None
            
        Returns:
            str: The unique name of this blueprint
        """
        return "decision_blueprint"
    
    @property
    def usage_context(self) -> str:
        """
        Description of when this blueprint should be used.
        
        Args:
            None
            
        Returns:
            str: Context description for when to use this blueprint
        """
        return ("Use this blueprint when the plan needs to make dynamic decisions based on tool outputs "
                "but cannot complete the full conditional logic in a single pass. This enables chaining "
                "multiple architect calls by storing intermediate results.")
    
    @property
    def purpose(self) -> str:
        """
        Description of what this blueprint is used for.
        
        Args:
            None
            
        Returns:
            str: Purpose description of what this blueprint accomplishes
        """
        return ("Stores intermediate calculation results and descriptions for conditional logic scenarios "
                "that require multiple architect calls to complete. Acts as a handoff mechanism between "
                "architect invocations. This blueprint SHOULD NOT be used unless absolutely necessary for conditional logic scenarios.")
    
    @property
    def parameter_instructions(self) -> Dict[str, Tuple[str, str]]:
        """
        Map of parameter names to data type and description tuples.
        
        Args:
            None
            
        Returns:
            Dict[str, Tuple[str, str]]: Map of parameter names to (DataType, description) tuples
        """
        return {
            "intermediate_results": (
                "List[str]", 
                "List of string descriptions of results followed by those results. For example: ['The result of tool_1 with inputs x and y was $ref.stage_1.tool_1.output_param_1', 'The result of tool_2 with inputs a and b was $ref.stage_1.tool_2.output_param_1']"
            )
        }
    
    @property
    def parameter_to_value_map(self) -> Dict[str, Any]:
        """
        Map of parameter names to their filled values.
        
        Args:
            None
            
        Returns:
            Dict[str, Any]: Map of parameter names to their current values
        """
        if not self._filled:
            return {}
        return {
            "intermediate_results": self._intermediate_results.copy()
        }
    
    def fill(self, parameters: Dict[str, Any]) -> None:
        """
        Fill in the blueprint with the provided parameters.
        
        Args:
            parameters (Dict[str, Any]): Dictionary containing 'intermediate_results' key with list of strings
            
        Returns:
            None: Fills the blueprint with provided parameters
            
        Raises:
            ValueError: If required parameters are missing or invalid
        """
        if "intermediate_results" not in parameters:
            raise ValueError("DecisionBlueprint requires 'intermediate_results' parameter")
        
        intermediate_results = parameters["intermediate_results"]
        
        if not isinstance(intermediate_results, list):
            raise ValueError("intermediate_results must be a list of strings")
        for item_index, item in enumerate(intermediate_results):
            if not isinstance(item, str):
                raise ValueError(f"All intermediate_results must be strings, but item {item_index} is {type(item)}")
        self._intermediate_results = intermediate_results.copy()
        self._filled = True
    
    def get_intermediate_results(self) -> List[str]:
        """
        Get the stored intermediate results.
        
        Args:
            None
            
        Returns:
            List[str]: List of string descriptions of intermediate results
            
        Raises:
            RuntimeError: If blueprint hasn't been filled yet
        """
        if not self._filled:
            raise RuntimeError("DecisionBlueprint has not been filled yet")
        return self._intermediate_results.copy()
    
    def is_filled(self) -> bool:
        """
        Check if the blueprint has been filled with parameters.
        
        Args:
            None
            
        Returns:
            bool: True if blueprint has been filled, False otherwise
        """
        return self._filled
    
    def format_for_next_call(self) -> str:
        """
        Format the intermediate results as context for the next architect call.
        
        Args:
            None
            
        Returns:
            str: Formatted string suitable for additional_context_prompt
            
        Raises:
            RuntimeError: If blueprint hasn't been filled yet
        """
        if not self._filled:
            raise RuntimeError("DecisionBlueprint has not been filled yet")
        
        if not self._intermediate_results:
            return "Previous architect call completed with no intermediate results."
        
        context = "Previous architect call results:\n"
        for result_index, result in enumerate(self._intermediate_results, 1):
            context += f"{result_index}. {result}\n"
        
        context += "\nUse these results to continue with the conditional logic that was requested."
        return context
